#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import warnings
from collections import OrderedDict
from collections.abc import Sequence
from copy import deepcopy
from dataclasses import dataclass, field
from functools import cached_property
from logging import Logger
from typing import Any, cast, Mapping

import torch
from ax.core.search_space import SearchSpaceDigest
from ax.exceptions.core import AxWarning, UnsupportedError, UserInputError
from ax.generators.torch_base import TorchOptConfig
from ax.generators.types import TConfig
from ax.utils.common.constants import Keys
from ax.utils.common.logger import get_logger
from ax.utils.common.typeutils import _argparse_type_encoder
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.logei import (
    qLogNoisyExpectedImprovement,
    qLogProbabilityOfFeasibility,
)
from botorch.acquisition.multi_objective.logei import (
    qLogNoisyExpectedHypervolumeImprovement,
)
from botorch.fit import fit_fully_bayesian_model_nuts, fit_gpytorch_mll
from botorch.models import PairwiseLaplaceMarginalLogLikelihood
from botorch.models.fully_bayesian import (
    AbstractFullyBayesianSingleTaskGP,
    FullyBayesianSingleTaskGP,
)
from botorch.models.fully_bayesian_multitask import SaasFullyBayesianMultiTaskGP
from botorch.models.gp_regression import SingleTaskGP
from botorch.models.gp_regression_fidelity import SingleTaskMultiFidelityGP
from botorch.models.gp_regression_mixed import MixedSingleTaskGP
from botorch.models.gpytorch import BatchedMultiOutputGPyTorchModel, GPyTorchModel
from botorch.models.model import Model, ModelList
from botorch.models.multitask import MultiTaskGP
from botorch.models.pairwise_gp import PairwiseGP
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform
from botorch.utils.constraints import get_outcome_constraint_transforms
from botorch.utils.datasets import RankingDataset, SupervisedDataset
from botorch.utils.dispatcher import Dispatcher
from botorch.utils.types import _DefaultType, DEFAULT
from gpytorch.kernels.kernel import Kernel
from gpytorch.likelihoods import Likelihood
from gpytorch.mlls.exact_marginal_log_likelihood import ExactMarginalLogLikelihood
from gpytorch.mlls.marginal_log_likelihood import MarginalLogLikelihood
from pyre_extensions import assert_is_instance, none_throws
from torch import Tensor


MIN_OBSERVED_NOISE_LEVEL = 1e-7
logger: Logger = get_logger(__name__)


@dataclass
class ModelConfig:
    """Configuration for the BoTorch Model used in Surrogate.

    Args:
        botorch_model_class: ``Model`` class to be used as the underlying
            BoTorch model. If None is provided a model class will be selected (either
            one for all outcomes or a ModelList with separate models for each outcome)
            will be selected automatically based off the datasets at `construct` time.
        model_options: Dictionary of options / kwargs for the BoTorch
            ``Model`` constructed during ``Surrogate.fit``.
            Note that the corresponding attribute will later be updated to include any
            additional kwargs passed into ``BoTorchGenerator.fit``.
        mll_class: ``MarginalLogLikelihood`` class to use for model-fitting.
        mll_options: Dictionary of options / kwargs for the MLL.
        outcome_transform_classes: List of BoTorch outcome transforms classes. Passed
            down to the BoTorch ``Model``. Multiple outcome transforms can be chained
            together using ``ChainedOutcomeTransform``.
        outcome_transform_options: Outcome transform classes kwargs. The keys are
            class string names and the values are dictionaries of outcome transform
            kwargs. For example,
            `
            outcome_transform_classes = [Standardize]
            outcome_transform_options = {
                "Standardize": {"m": 1},
            `
            For more options see `botorch/models/transforms/outcome.py`.
        input_transform_classes: List of BoTorch input transforms classes.
            Passed down to the BoTorch ``Model``. Multiple input transforms
            will be chained together using ``ChainedInputTransform``.
            If `DEFAULT`, a default set of input transforms may be constructed
            based on the search space digest (in `_construct_default_input_transforms`).
            To disable this behavior, pass in `input_transform_classes=None`.
        input_transform_options: Input transform classes kwargs. The keys are
            class string names and the values are dictionaries of input transform
            kwargs. For example,
            `
            input_transform_classes = [Normalize, Round]
            input_transform_options = {
                "Normalize": {"d": 3},
                "Round": {"integer_indices": [0], "categorical_features": {1: 2}},
            }
            `
            For more input options see `botorch/models/transforms/input.py`.
        covar_module_class: Covariance module class. This gets initialized after
            parsing the ``covar_module_options`` in ``covar_module_argparse``,
            and gets passed to the model constructor as ``covar_module``.
        covar_module_options: Covariance module kwargs.
        likelihood_class: ``Likelihood`` class. This gets initialized with
            ``likelihood_options`` and gets passed to the model constructor.
        likelihood_options: Likelihood options.
        name: Name of the model config. This is used to identify the model config.
    """

    botorch_model_class: type[Model] | None = None
    model_options: dict[str, Any] = field(default_factory=dict)
    mll_class: type[MarginalLogLikelihood] | None = None
    mll_options: dict[str, Any] = field(default_factory=dict)
    input_transform_classes: list[type[InputTransform]] | _DefaultType | None = DEFAULT
    input_transform_options: dict[str, dict[str, Any]] | None = field(
        default_factory=dict
    )
    outcome_transform_classes: list[type[OutcomeTransform]] | None = None
    outcome_transform_options: dict[str, dict[str, Any]] = field(default_factory=dict)
    covar_module_class: type[Kernel] | None = None
    covar_module_options: dict[str, Any] = field(default_factory=dict)
    likelihood_class: type[Likelihood] | None = None
    likelihood_options: dict[str, Any] = field(default_factory=dict)
    name: str | None = None

    @cached_property
    def identifier(self) -> str:
        """Returns a unique identifier for the model config."""
        return self.name if self.name is not None else str(self)


def use_model_list(
    datasets: Sequence[SupervisedDataset],
    search_space_digest: SearchSpaceDigest,
    model_configs: list[ModelConfig] | None = None,
    metric_to_model_configs: dict[str, list[ModelConfig]] | None = None,
    allow_batched_models: bool = True,
) -> bool:
    model_configs = model_configs or []
    metric_to_model_configs = metric_to_model_configs or {}
    if len(datasets) == 1 and datasets[0].Y.shape[-1] == 1:
        # There is only one outcome, so we can use a single model.
        return False
    elif (
        len(model_configs) > 1
        or len(metric_to_model_configs) > 0
        or any(len(model_config) for model_config in metric_to_model_configs.values())
    ):
        # There are multiple outcomes and outcomes might be modeled with different
        # models
        return True
    if len({type(d) for d in datasets}) > 1:
        # Use a `ModelList` if there are multiple dataset classes.
        return True

    botorch_model_class_set = {mc.botorch_model_class for mc in model_configs}
    # if any of the botorch_model_class is unspecified, we'd need to infer its class
    if (not botorch_model_class_set) or (None in botorch_model_class_set):
        inferred_botorch_model_class_set = {
            choose_model_class(dataset=dataset, search_space_digest=search_space_digest)
            for dataset in datasets
        }
        botorch_model_class_set = botorch_model_class_set.union(
            inferred_botorch_model_class_set
        )
        # Safe even if None is not present
        botorch_model_class_set.discard(None)

    if len(botorch_model_class_set) > 1:
        # Use a `ModelList` if there are multiple possible botorch_model_class classes.
        return True

    # Otherwise, the same model class is used for all outcomes.
    botorch_model_class = none_throws(next(iter(botorch_model_class_set)))
    if issubclass(botorch_model_class, FullyBayesianSingleTaskGP):
        # SAAS models do not support multiple outcomes.
        # Use model list if there are multiple outcomes.
        return len(datasets) > 1 or datasets[0].Y.shape[-1] > 1
    elif issubclass(botorch_model_class, MultiTaskGP):
        # We wrap multi-task models into `ModelListGP` when there are
        # multiple outcomes.
        return len(datasets) > 1 or datasets[0].Y.shape[-1] > 1
    elif len(datasets) == 1:
        # This method is called before multiple datasets are merged into
        # one if using a batched model. If there is one dataset here,
        # there should be a reason that a single model should be used:
        # e.g. a contextual model, where we want to jointly model the metric
        # each context (and context-level metrics are different outcomes).
        return False
    elif issubclass(botorch_model_class, BatchedMultiOutputGPyTorchModel) and all(
        torch.equal(datasets[0].X, ds.X) for ds in datasets[1:]
    ):
        # Use batch models if allowed
        return not allow_batched_models
    # If there are multiple Xs and they are not all equal, we use `ModelListGP`.
    return True


def copy_model_config_with_default_values(
    model_config: ModelConfig,
    dataset: SupervisedDataset,
    search_space_digest: SearchSpaceDigest,
) -> ModelConfig:
    model_config_copy = deepcopy(model_config)

    if model_config_copy.botorch_model_class is None:
        model_config_copy.botorch_model_class = choose_model_class(
            dataset=dataset, search_space_digest=search_space_digest
        )

    if model_config_copy.mll_class is None:
        model_config_copy.mll_class = (
            PairwiseLaplaceMarginalLogLikelihood
            if model_config_copy.botorch_model_class is PairwiseGP
            else ExactMarginalLogLikelihood
        )

    # PairwiseGP does not use outcome transforms
    if model_config_copy.outcome_transform_classes is not None:
        if model_config_copy.botorch_model_class is PairwiseGP:
            model_config_copy.outcome_transform_classes = None

    return model_config_copy


def choose_model_class(
    dataset: SupervisedDataset,
    search_space_digest: SearchSpaceDigest,
) -> type[Model]:
    """Chooses a BoTorch `Model` class and `MarginalLogLikelihood` class
    using the given the dataset and search_space_digest.

    Args:
        dataset: The dataset on which the model will be fitted.
        search_space_digest: The digest of the search space the model will be
            fitted within.

    Returns:
        A BoTorch `Model` class.
    """
    if len(search_space_digest.fidelity_features) > 1:
        raise NotImplementedError(
            "Only a single fidelity feature supported "
            f"(got: {search_space_digest.fidelity_features})."
        )
    if len(search_space_digest.task_features) > 1:
        raise NotImplementedError(
            f"Only a single task feature supported "
            f"(got: {search_space_digest.task_features})."
        )
    if search_space_digest.task_features and search_space_digest.fidelity_features:
        raise NotImplementedError(
            "Multi-task multi-fidelity optimization not yet supported."
        )

    # Preference learning case
    if isinstance(dataset, RankingDataset):
        model_class = PairwiseGP

    # Multi-task case (when `task_features` is specified).
    elif search_space_digest.task_features:
        model_class = MultiTaskGP

    # Single-task multi-fidelity cases.
    elif search_space_digest.fidelity_features:
        model_class = SingleTaskMultiFidelityGP

    # Mixed optimization case. Note that presence of categorical
    # features in search space digest indicates that downstream in the
    # stack we chose not to perform continuous relaxation on those
    # features.
    elif search_space_digest.categorical_features:
        model_class = MixedSingleTaskGP

    # Single-task single-fidelity cases.
    else:
        model_class = SingleTaskGP

    logger.debug(f"Chose BoTorch model class: {model_class}.")
    return model_class


def choose_botorch_acqf_class(
    torch_opt_config: TorchOptConfig,
    datasets: Sequence[SupervisedDataset] | None,
    use_p_feasible: bool = True,
) -> type[AcquisitionFunction]:
    """Chooses the most suitable BoTorch `AcquisitionFunction` class.

    Args:
        torch_opt_config: The torch optimization config.
        datasets: The datasets that were used to fit the model.
        use_p_feasible: Whether we dispatch to `qLogProbabilityOfFeasibility` when
            there are no feasible points in the training data.

    Returns:
        A BoTorch `AcquisitionFunction` class. The current logic chooses between:
            - `qLogProbabilityOfFeasibility` if there are outcome constraints and
                no feasible point has been found.
            - `qLogNoisyExpectedImprovement` for single-objective optimization.
            - `qLogNoisyExpectedHypervolumeImprovement`` for multi-objective
                optimization.
    """

    # Check if the training data is feasible.
    if (
        use_p_feasible
        and torch_opt_config.outcome_constraints is not None
        and datasets is not None
    ):
        con_tfs = (
            get_outcome_constraint_transforms(torch_opt_config.outcome_constraints)
            or []
        )
        # NOTE: `convert_to_block_design` will drop points that are only observed by
        # some of the metrics which is natural as we are using observed values to
        # determine feasibility.
        dataset = convert_to_block_design(datasets=datasets, force=True)[0]
        con_observed = torch.stack([con(dataset.Y) for con in con_tfs], dim=-1)
        feas_point_found = (con_observed <= 0).all(dim=-1).any().item()

        if not feas_point_found:
            acqf_class = qLogProbabilityOfFeasibility
            logger.debug(f"Chose BoTorch acquisition function class: {acqf_class}.")
            return acqf_class

    if torch_opt_config.is_moo and not torch_opt_config.use_learned_objective:
        acqf_class = qLogNoisyExpectedHypervolumeImprovement
    else:
        acqf_class = qLogNoisyExpectedImprovement

    logger.debug(f"Chose BoTorch acquisition function class: {acqf_class}.")
    return acqf_class


def construct_acquisition_and_optimizer_options(
    acqf_options: TConfig,
    botorch_acqf_options: TConfig,
    model_gen_options: TConfig | None = None,
    botorch_acqf_classes_with_options: list[tuple[type[AcquisitionFunction], TConfig]]
    | None = None,
) -> tuple[
    TConfig, TConfig, TConfig, list[tuple[type[AcquisitionFunction], TConfig]] | None
]:
    """Extract acquisition and optimizer options from `model_gen_options`."""
    acq_options = acqf_options.copy()
    botorch_acqf_options = botorch_acqf_options.copy()
    opt_options = {}

    if model_gen_options:
        # Define the allowed paths

        if (
            len(
                extra_keys_in_model_gen_options := set(model_gen_options.keys())
                - {
                    Keys.OPTIMIZER_KWARGS.value,
                    Keys.ACQF_KWARGS.value,
                    Keys.AX_ACQUISITION_KWARGS.value,
                }
            )
            > 0
        ):
            raise ValueError(
                "Found forbidden keys in `model_gen_options`: "
                f"{extra_keys_in_model_gen_options}."
            )
        new_botorch_acqf_options = assert_is_instance(
            model_gen_options.get(Keys.ACQF_KWARGS, {}),
            dict,
        )
        if (
            # pyre-fixme [6]: Incompatible parameter type [6]: In call `len`,
            # for 1st positional argument, expected
            # `pyre_extensions.PyreReadOnly[Sized]` but got `dict`.
            len(new_botorch_acqf_options) > 0
            and botorch_acqf_classes_with_options is not None
        ):
            if len(botorch_acqf_classes_with_options) > 1:
                warnings.warn(
                    message="botorch_acqf_options are being ignored, due to using "
                    "MultiAcquisition. Specify options for each acquistion function"
                    "via botorch_acqf_classes_with_options.",
                    category=AxWarning,
                    stacklevel=4,
                )
            else:
                botorch_acqf_classes_with_options = deepcopy(
                    botorch_acqf_classes_with_options
                )
                botorch_acqf_classes_with_options[0][1].update(new_botorch_acqf_options)
        else:
            botorch_acqf_options.update(new_botorch_acqf_options)

        acq_options.update(
            assert_is_instance(
                model_gen_options.get(Keys.AX_ACQUISITION_KWARGS, {}),
                dict,
            )
        )

        # TODO: Add this if all acq. functions accept the `subset_model`
        # kwarg or opt for kwarg filtering.
        # acq_options[SUBSET_MODEL] = model_gen_options.get(SUBSET_MODEL)
        opt_options = assert_is_instance(
            model_gen_options.get(Keys.OPTIMIZER_KWARGS, {}),
            dict,
        ).copy()
    return (
        acq_options,
        botorch_acqf_options,
        opt_options,
        botorch_acqf_classes_with_options,
    )


def convert_to_block_design(
    datasets: Sequence[SupervisedDataset],
    force: bool = False,
) -> list[SupervisedDataset]:
    """Converts a list of datasets to a single block-design dataset that contains
    all outcomes.

    Args:
        datasets: A list of datasets to merge.
        force: If True, will force conversion of data not complying to a block
            design to block design by dropping observations that are not shared
            between outcomes.
            If only a subset of the outcomes have noise observations, all noise
            observations will be dropped.

    Returns:
        A single element list containing the merged dataset.
    """
    is_fixed = [ds.Yvar is not None for ds in datasets]
    if any(is_fixed) and not all(is_fixed):
        if force:
            logger.debug(
                "Only a subset of datasets have noise observations. "
                "Dropping all noise observations since `force=True`. "
            )
        else:
            raise UnsupportedError(
                "Cannot convert mixed data with and without variance "
                "observations to `block design`."
            )
    is_fixed = all(is_fixed)
    Xs = [dataset.X for dataset in datasets]
    for dset in datasets[1:]:
        if dset.feature_names != datasets[0].feature_names:
            raise ValueError(
                "Feature names must be the same across all datasets, "
                f"got {dset.feature_names} and {datasets[0].feature_names}"
            )

    # Join the outcome names of datasets.
    outcome_names = sum([ds.outcome_names for ds in datasets], [])

    if len({X.shape for X in Xs}) != 1 or not all(
        torch.equal(X, Xs[0]) for X in Xs[1:]
    ):
        if not force:
            raise UnsupportedError(
                "Cannot convert data to non-block design data. "
                "To force this and drop data not shared between "
                "outcomes use `force=True`."
            )
        logger.debug(
            "Forcing conversion of data not complying to a block design "
            "to block design by dropping observations that are not shared "
            "between outcomes."
        )
        X_shared, idcs_shared = _get_shared_rows(Xs=Xs)
        Y = torch.cat([ds.Y[i] for ds, i in zip(datasets, idcs_shared)], dim=-1)
        if is_fixed:
            Yvar = torch.cat(
                # pyre-fixme[16]: `Optional` has no attribute `__getitem__`.
                [ds.Yvar[i] for ds, i in zip(datasets, idcs_shared)],
                dim=-1,
            )
        else:
            Yvar = None
        datasets = [
            SupervisedDataset(
                X=X_shared,
                Y=Y,
                Yvar=Yvar,
                feature_names=datasets[0].feature_names,
                outcome_names=outcome_names,
            )
        ]
        return datasets

    # data complies to block design, can concat with impunity
    Y = torch.cat([ds.Y for ds in datasets], dim=-1)
    if is_fixed:
        Yvar = torch.cat([none_throws(ds.Yvar) for ds in datasets], dim=-1)
    else:
        Yvar = None
    datasets = [
        SupervisedDataset(
            X=Xs[0],
            Y=Y,
            Yvar=Yvar,
            feature_names=datasets[0].feature_names,
            outcome_names=outcome_names,
        )
    ]
    return datasets


def _get_shared_rows(Xs: list[Tensor]) -> tuple[Tensor, list[Tensor]]:
    """Extract shared rows from a list of tensors

    Args:
        Xs: A list of m two-dimensional tensors with shapes
            `(n_1 x d), ..., (n_m x d)`. It is not required that
            the `n_i` are the same.

    Returns:
        A two-tuple containing (i) a Tensor with the rows that are
        shared between all the Tensors in `Xs`, and (ii) a list of
        index tensors that indicate the location of these rows
        in the respective elements of `Xs`.
    """
    if any(X.ndim != 2 for X in Xs):
        raise UserInputError("All inputs must be two-dimensional.")
    idcs_shared = []
    Xs_sorted = sorted(Xs, key=len)
    X_shared = Xs_sorted[0].clone().unique(dim=0)
    for X in Xs_sorted[1:]:
        X_shared = X_shared[(X_shared == X.unsqueeze(-2)).all(dim=-1).any(dim=-2)]
    # get indices
    for X in Xs:
        _, inverse_indices = X.unique(dim=0, return_inverse=True)
        # set of unique original indices
        unique_indices = set(
            {v: i for i, v in enumerate(inverse_indices.tolist())}.values()
        )
        same = (X_shared == X.unsqueeze(-2)).all(dim=-1).any(dim=-1)
        shared_idcs_list = []
        for i, is_shared in enumerate(same):
            if is_shared and i in unique_indices:
                shared_idcs_list.append(i)
        shared_idcs = torch.tensor(
            shared_idcs_list, dtype=torch.long, device=X_shared.device
        )
        idcs_shared.append(shared_idcs)
    return X_shared, idcs_shared


def subset_state_dict(
    state_dict: Mapping[str, Tensor],
    submodel_index: int,
) -> OrderedDict[str, Tensor]:
    """Get the state dict for a submodel from the state dict of a model list.

    Args:
        state_dict: A state dict.
        submodel_index: The index of the submodel to extract.

    Returns:
        The state dict for the submodel.
    """
    expected_substring = f"models.{submodel_index}."
    len_substring = len(expected_substring)
    new_items = [
        (k[len_substring:], v)
        for k, v in state_dict.items()
        if k.startswith(expected_substring)
    ]
    return OrderedDict(new_items)


# ----------------------- Model fitting helpers ----------------------- #


fit_botorch_model = Dispatcher(name="fit_botorch_model", encoder=_argparse_type_encoder)


@fit_botorch_model.register(ModelList)
def _fit_botorch_model_list(
    model: Model,
    mll_class: type[MarginalLogLikelihood],
    mll_options: dict[str, Any] | None = None,
) -> None:
    for m in cast(list[Model], model.models):
        fit_botorch_model(m, mll_class=mll_class, mll_options=mll_options)


@fit_botorch_model.register(GPyTorchModel)
@fit_botorch_model.register(PairwiseGP)
def _fit_botorch_model_gpytorch(
    model: GPyTorchModel | PairwiseGP,
    mll_class: type[MarginalLogLikelihood],
    mll_options: dict[str, Any] | None = None,
) -> None:
    """Fit a GPyTorch based BoTorch model."""
    mll_options = mll_options or {}
    mll = mll_class(likelihood=model.likelihood, model=model, **mll_options)
    fit_gpytorch_mll(mll)


@fit_botorch_model.register(
    (AbstractFullyBayesianSingleTaskGP, SaasFullyBayesianMultiTaskGP)
)
def _fit_botorch_model_fully_bayesian_nuts(
    model: AbstractFullyBayesianSingleTaskGP | SaasFullyBayesianMultiTaskGP,
    mll_class: type[MarginalLogLikelihood],
    mll_options: dict[str, Any] | None = None,
) -> None:
    mll_options = mll_options or {}
    mll_options.setdefault("disable_progbar", True)
    fit_fully_bayesian_model_nuts(model, **mll_options)


@fit_botorch_model.register(object)
def _fit_botorch_model_not_implemented(
    model: Model,
    mll_class: type[MarginalLogLikelihood],
    mll_options: dict[str, Any] | None = None,
) -> None:
    raise NotImplementedError(
        f"fit_botorch_model is not implemented for {model.__class__.__name__}. "
        "You can register a model fitting routine for it by adding new case "
        "to the `fit_botorch_model` dispatcher. To do so, decorate a function "
        "that accepts `model`, `mll_class` and `mll_options` inputs with "
        f"`@fit_botorch_model.register({model.__class__.__name__})`."
    )


@dataclass(frozen=True)
class CVFold:
    """
    Args:
        train_dataset: The training dataset for the fold.
        test_X: The test inputs for the fold.
        test_Y: The test outputs for the fold.
    """

    train_dataset: SupervisedDataset
    test_X: Tensor
    test_Y: Tensor


def get_cv_fold(
    dataset: SupervisedDataset, X: Tensor, Y: Tensor, idcs: Tensor
) -> CVFold:
    train_mask = torch.ones(X.shape[0], dtype=torch.bool, device=X.device)
    train_mask[idcs] = 0
    return CVFold(
        train_dataset=dataset.clone(mask=train_mask),
        test_X=X[idcs],
        test_Y=Y[idcs],
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from logging import Logger
from typing import Any, TYPE_CHECKING

from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.metadata_to_parameter import MetadataToParameterMixin
from ax.core import ParameterType
from ax.core.observation import Observation
from ax.core.parameter import ChoiceParameter, TParamValue
from ax.core.search_space import SearchSpace
from ax.generators.types import TConfig
from ax.utils.common.constants import Keys
from ax.utils.common.logger import get_logger
from pyre_extensions import assert_is_instance

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


logger: Logger = get_logger(__name__)


class MetadataToTask(MetadataToParameterMixin, Transform):
    """
    This transform converts metadata from observation features into a task parameter.

    This is used in transfer learning to specify the task for each observation.

    It allows the user to specify the `config` with `task_values` as the key, which
    contains the unique task values for the search space.

    Transform is done in-place.
    """

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: adapter_module.base.Adapter | None = None,
        config: TConfig | None = None,
    ) -> None:
        Transform.__init__(
            self,
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        config = config or {}
        task_values: list[TParamValue] = [
            assert_is_instance(v, TParamValue)
            for v in assert_is_instance(config["task_values"], list)
        ]
        self.parameters: dict[str, dict[str, Any]] = {Keys.TASK_FEATURE_NAME.value: {}}
        self._parameter_list = [
            ChoiceParameter(
                name=Keys.TASK_FEATURE_NAME.value,
                parameter_type=ParameterType.INT,
                values=task_values,
                is_task=True,
                target_value=0,
                is_ordered=False,
                sort_values=True,
            )
        ]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from collections.abc import Callable
from enum import Enum, unique
from importlib import import_module
from typing import Any

from ax.exceptions.core import UnsupportedError


@unique
class FuncEnum(Enum):
    """A base class for all enums with the following structure: string values that
    map to names of functions, which reside in the same module as the enum."""

    # pyre-ignore[3]: Input constructors will be used to make different inputs,
    # so we need to allow `Any` return type here.
    def __call__(self, **kwargs: Any) -> Any:
        """Defines a method, by which the members of this enum can be called,
        e.g. ``MyFunctions.F(**kwargs)``, which will call the corresponding
        function registered by the name ``F`` in the enum."""
        return self._get_function_for_value()(**kwargs)

    # pyre-ignore[31]: Expression `typing.Callable[([...], typing.Any)]`
    # is not a valid type.
    def _get_function_for_value(self) -> Callable[[...], Any]:
        """Retrieve the function in this module, name of which corresponds to the
        value of the enum member."""
        try:
            return getattr(import_module(self.__module__), self.value)
        except AttributeError:
            raise UnsupportedError(
                f"{self.value} is not defined as a method in "
                f"`{self.__module__}`. Please add the method "
                "to the file."
            )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy.typing as npt
from ax.metrics.noisy_function import NoisyFunctionMetric
from ax.utils.measurement.synthetic_functions import aug_hartmann6, hartmann6
from pyre_extensions import assert_is_instance


class Hartmann6Metric(NoisyFunctionMetric):
    def f(self, x: npt.NDArray) -> float:
        return assert_is_instance(hartmann6(x), float)


class AugmentedHartmann6Metric(NoisyFunctionMetric):
    def f(self, x: npt.NDArray) -> float:
        return assert_is_instance(aug_hartmann6(x), float)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, Union

import numpy as np
import numpy.typing as npt
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generator_spec import GeneratorSpec
from ax.utils.common.base import Base
from ax.utils.common.func_enum import FuncEnum
from pyre_extensions import none_throws

# pyre-fixme[24]: Generic type `np.ndarray` expects 2 type parameters.
ARRAYLIKE = Union[np.ndarray, list[float], list[np.ndarray]]


class BestModelSelector(ABC, Base):
    @abstractmethod
    def best_model(self, generator_specs: list[GeneratorSpec]) -> GeneratorSpec:
        """Return the best ``GeneratorSpec`` based on some criteria.

        NOTE: The returned ``GeneratorSpec`` may be a different object than
        what was provided in the original list. It may be possible to
        clone and modify the original ``GeneratorSpec`` to produce one that
        performs better.
        """


# Functions used in ReductionCriterion. They're looked up from the same module.
mean: Callable[[ARRAYLIKE], npt.NDArray] = np.mean
min: Callable[[ARRAYLIKE], npt.NDArray] = np.min
max: Callable[[ARRAYLIKE], npt.NDArray] = np.max


class ReductionCriterion(FuncEnum):
    """An enum for callables that are used for aggregating diagnostics over metrics
    and selecting the best diagnostic in ``SingleDiagnosticBestModelSelector``.

    NOTE: This is used to ensure serializability of the callables.
    """

    MEAN = "mean"
    MIN = "min"
    MAX = "max"

    def __call__(self, array_like: ARRAYLIKE) -> npt.NDArray:
        """This will look up the corresponding function in the same module and
        call it with the given arguments.
        """
        return self._get_function_for_value()(array_like)


class SingleDiagnosticBestModelSelector(BestModelSelector):
    """Choose the best model using a single cross-validation diagnostic.

    The input is a list of ``GeneratorSpec``, each corresponding to one model.
    The specified diagnostic is extracted from each of the models,
    its values (each of which corresponds to a separate metric) are
    aggregated with the aggregation function, the best one is determined
    with the criterion, and the index of the best diagnostic result is returned.

    Example:
     ::
        s = SingleDiagnosticBestModelSelector(
            diagnostic='Fisher exact test p',
            metric_aggregation=ReductionCriterion.MEAN,
            criterion=ReductionCriterion.MIN,
            model_cv_kwargs={"untransform": False},
        )
        best_model = s.best_model(generator_specs=generator_specs)

    Args:
        diagnostic: The name of the diagnostic to use, which should be
            a key in ``CVDiagnostic``.
        metric_aggregation: ``ReductionCriterion`` applied to the values of the
            diagnostic for a single model to produce a single number.
        criterion: ``ReductionCriterion`` used to determine which of the
            (aggregated) diagnostics is the best.
        model_cv_kwargs: Optional dictionary of kwargs to pass in while computing
            the cross validation diagnostics.
    """

    def __init__(
        self,
        diagnostic: str,
        metric_aggregation: ReductionCriterion,
        criterion: ReductionCriterion,
        model_cv_kwargs: dict[str, Any] | None = None,
    ) -> None:
        self.diagnostic = diagnostic
        if not isinstance(metric_aggregation, ReductionCriterion) or not isinstance(
            criterion, ReductionCriterion
        ):
            raise UserInputError(
                "Both `metric_aggregation` and `criterion` must be "
                f"`ReductionCriterion`. Got {metric_aggregation=}, {criterion=}."
            )
        if criterion == ReductionCriterion.MEAN:
            raise UserInputError(
                f"{criterion=} is not supported. Please use MIN or MAX."
            )
        self.metric_aggregation = metric_aggregation
        self.criterion = criterion
        self.model_cv_kwargs = model_cv_kwargs

    def best_model(self, generator_specs: list[GeneratorSpec]) -> GeneratorSpec:
        """Return the best ``GeneratorSpec`` based on the specified diagnostic.

        Args:
            generator_specs: List of ``GeneratorSpec`` to choose from.

        Returns:
            The best ``GeneratorSpec`` based on the specified diagnostic.
        """
        for generator_spec in generator_specs:
            generator_spec.cross_validate(model_cv_kwargs=self.model_cv_kwargs)
        aggregated_diagnostic_values = [
            self.metric_aggregation(
                list(none_throws(generator_spec.diagnostics)[self.diagnostic].values())
            )
            for generator_spec in generator_specs
        ]
        best_diagnostic = self.criterion(aggregated_diagnostic_values).item()
        best_index = aggregated_diagnostic_values.index(best_diagnostic)
        return generator_specs[best_index]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from typing import Any, Literal

from ax.api.types import TParameterValue
from ax.storage.registry_bundle import RegistryBundleBase


@dataclass
class RangeParameterConfig:
    """
    Allows specifying a continuous dimension of an experiment's search space
    and internally validates the inputs.
    """

    name: str

    bounds: tuple[float, float]
    parameter_type: Literal["float", "int"]
    step_size: float | None = None
    scaling: Literal["linear", "log"] | None = None


@dataclass
class ChoiceParameterConfig:
    """
    Allows specifying a discrete dimension of an experiment's search space and
    internally validates the inputs. Choice parameters can be either ordinal or
    categorical; this is controlled via the ``is_ordered`` flag.
    """

    name: str
    values: list[float] | list[int] | list[str] | list[bool]
    parameter_type: Literal["float", "int", "str", "bool"]
    is_ordered: bool | None = None
    dependent_parameters: Mapping[TParameterValue, Sequence[str]] | None = None


@dataclass
class StorageConfig:
    """
    Allows the user to configure how Ax should connect to a SQL database to store the
    experiment and its data.
    """

    creator: Callable[..., Any] | None = None
    url: str | None = None
    registry_bundle: RegistryBundleBase | None = None



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Sized
from copy import deepcopy

import numpy as np
from ax.adapter.base import DataLoaderConfig
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.transforms.choice_encode import (
    ChoiceToNumericChoice,
    OrderedChoiceEncode,
    OrderedChoiceToIntegerRange,
)
from ax.core.observation import ObservationFeatures
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.parameter_constraint import ParameterConstraint
from ax.core.search_space import HierarchicalSearchSpace, RobustSearchSpace, SearchSpace
from ax.core.types import TParameterization
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_experiment_with_observations,
    get_robust_search_space,
)
from pandas import DataFrame
from pandas.testing import assert_frame_equal
from pyre_extensions import assert_is_instance


class ChoiceEncodeTransformTest(TestCase):
    t_class = ChoiceToNumericChoice

    def setUp(self) -> None:
        super().setUp()
        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "x", lower=1, upper=3, parameter_type=ParameterType.FLOAT
                ),
                RangeParameter("a", lower=1, upper=2, parameter_type=ParameterType.INT),
                ChoiceParameter(
                    "b",
                    parameter_type=ParameterType.FLOAT,
                    values=[1.0, 10.0, 100.0],
                    is_ordered=True,
                ),
                ChoiceParameter(
                    "c",
                    parameter_type=ParameterType.FLOAT,
                    values=[10.0, 100.0, 1000.0],
                    is_ordered=True,
                    sort_values=False,
                ),
                ChoiceParameter(
                    "d",
                    parameter_type=ParameterType.STRING,
                    values=["r", "q", "z"],
                    sort_values=True,
                ),
                ChoiceParameter(
                    "e",
                    parameter_type=ParameterType.STRING,
                    values=["r", "q", "z"],
                    is_ordered=False,
                    sort_values=False,
                ),
            ],
            parameter_constraints=[
                ParameterConstraint(constraint_dict={"x": -0.5, "a": 1}, bound=0.5)
            ],
        )
        self.t = self.t_class(search_space=self.search_space)
        input_params: TParameterization = {
            "x": 2.2,
            "a": 2,
            "b": 10.0,
            "c": 10.0,
            "d": "r",
        }
        self.observation_features = [ObservationFeatures(parameters=input_params)]
        # expected parameters after transform
        self.expected_transformed_params: TParameterization = {
            **input_params,
            # string choice originally; transformed to int index.
            **{"d": 1},
        }

    def test_init(self) -> None:
        self.assertEqual(list(self.t.encoded_parameters.keys()), ["d", "e"])

    def test_transform_observation_features(self) -> None:
        observation_features = self.observation_features
        obs_ft2 = deepcopy(observation_features)
        obs_ft2 = self.t.transform_observation_features(obs_ft2)
        self.assertEqual(
            obs_ft2,
            [ObservationFeatures(parameters=self.expected_transformed_params)],
        )
        obs_ft2 = self.t.untransform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, observation_features)
        # Test transform on partial features
        obs_ft3 = [ObservationFeatures(parameters={"x": 2.2, "b": 10.0})]
        obs_ft3 = self.t.transform_observation_features(obs_ft3)
        self.assertEqual(
            obs_ft3[0],
            ObservationFeatures(
                parameters={"x": 2.2, "b": self.expected_transformed_params["b"]}
            ),
        )
        obs_ft5 = self.t.transform_observation_features([ObservationFeatures({})])
        self.assertEqual(obs_ft5[0], ObservationFeatures({}))

    def test_parameter_attributes_are_preserved(self) -> None:
        ss2 = deepcopy(self.search_space)
        ss2 = self.t.transform_search_space(ss2)
        for p in ("d", "e"):
            with self.subTest(p):
                transformed_param = assert_is_instance(
                    ss2.parameters[p], ChoiceParameter
                )
                original_param = assert_is_instance(
                    self.search_space.parameters[p], ChoiceParameter
                )
                self.assertEqual(
                    transformed_param.is_ordered, original_param.is_ordered
                )
                self.assertEqual(
                    transformed_param.sort_values,
                    original_param.sort_values,
                )
                if self.t_class == ChoiceToNumericChoice:
                    self.assertEqual(
                        transformed_param.values,
                        [i for i, _ in enumerate(original_param.values)],
                    )
                else:
                    self.assertEqual(
                        transformed_param.values,
                        original_param.values,
                    )

    def test_transform_search_space(self) -> None:
        ss2 = deepcopy(self.search_space)
        ss2 = self.t.transform_search_space(ss2)

        for p in ("x", "a"):
            self.assertIsInstance(ss2.parameters[p], RangeParameter)
        for p in ("b", "c", "d"):
            self.assertIsInstance(ss2.parameters[p], ChoiceParameter)
        for p in ("x", "b", "c"):
            self.assertEqual(ss2.parameters[p].parameter_type, ParameterType.FLOAT)
        for p in ("a", "d"):
            self.assertEqual(ss2.parameters[p].parameter_type, ParameterType.INT)

        for param_name in ["b", "c"]:
            self.assertEqual(
                ss2.parameters[param_name].values,
                assert_is_instance(
                    self.search_space[param_name], ChoiceParameter
                ).values,
            )
        self.assertEqual(ss2.parameters["d"].values, [0, 1, 2])

        # Fidelity parameter is transformed correctly.
        ss3 = SearchSpace(
            parameters=[
                ChoiceParameter(
                    "b",
                    parameter_type=ParameterType.STRING,
                    values=["a", "b", "c"],
                    is_ordered=True,
                    is_fidelity=True,
                    sort_values=False,
                    target_value="c",
                )
            ]
        )
        t = ChoiceToNumericChoice(search_space=ss3, observations=[])
        self.assertEqual(
            t.transform_search_space(ss3.clone()),
            SearchSpace(
                parameters=[
                    ChoiceParameter(
                        "b",
                        parameter_type=ParameterType.INT,
                        values=[0, 1, 2],
                        is_ordered=True,
                        is_fidelity=True,
                        sort_values=False,
                        target_value=2,
                    )
                ]
            ),
        )

    def test_hss_dependents_are_preserved(self) -> None:
        # x0
        # ├── x1
        # └── x2
        #     ├── (False) EMPTY
        #     └── (True) x3
        hss = HierarchicalSearchSpace(
            parameters=[
                FixedParameter(
                    "x0",
                    parameter_type=ParameterType.BOOL,
                    value=True,
                    dependents={True: ["x1", "x2"]},
                ),
                RangeParameter(
                    "x1",
                    parameter_type=ParameterType.FLOAT,
                    lower=0.0,
                    upper=1.0,
                ),
                ChoiceParameter(
                    "x2",
                    parameter_type=ParameterType.STRING,
                    values=["NO", "YES"],
                    is_ordered=True,
                    sort_values=True,
                    dependents={"NO": [], "YES": ["x3"]},
                ),
                RangeParameter(
                    "x3",
                    parameter_type=ParameterType.FLOAT,
                    lower=0.0,
                    upper=1.0,
                ),
            ]
        )
        hss = self.t_class(search_space=hss).transform_search_space(hss)

        # x0 should be untouched because it's a fixed parameter.
        self.assertIsInstance(hss.parameters["x0"], FixedParameter)
        self.assertEqual(hss.parameters["x0"].parameter_type, ParameterType.BOOL)
        # pyre-ignore[16] # Pyre doesn't understand fixed parameters have `.value`
        self.assertEqual(hss.parameters["x0"].value, True)
        self.assertEqual(hss.parameters["x0"].dependents, {True: ["x1", "x2"]})

        self.assertFalse(hss.parameters["x1"].is_hierarchical)
        self.assertFalse(hss.parameters["x3"].is_hierarchical)

        self.assertTrue(hss.parameters["x2"].is_hierarchical)
        self.assertEqual(hss.parameters["x2"].parameter_type, ParameterType.INT)
        self.assertEqual(hss.parameters["x2"].dependents, {0: [], 1: ["x3"]})

    def test_with_parameter_distributions(self) -> None:
        rss = get_robust_search_space()
        assert_is_instance(rss.parameters["c"], ChoiceParameter)._is_ordered = True
        # Transform a non-distributional parameter.
        t = self.t_class(
            search_space=rss,
            observations=[],
        )
        rss_new = assert_is_instance(t.transform_search_space(rss), RobustSearchSpace)
        self.assertEqual(set(rss.parameters.keys()), set(rss_new.parameters.keys()))
        self.assertEqual(rss.parameter_distributions, rss_new.parameter_distributions)
        self.assertEqual(rss_new.parameters["c"].parameter_type, ParameterType.INT)
        # Test with environmental variables.
        all_params = list(rss.parameters.values())
        rss = RobustSearchSpace(
            parameters=all_params[2:],
            parameter_distributions=rss.parameter_distributions,
            num_samples=rss.num_samples,
            environmental_variables=all_params[:2],
        )
        t = self.t_class(
            search_space=rss,
            observations=[],
        )
        rss_new = assert_is_instance(t.transform_search_space(rss), RobustSearchSpace)
        self.assertEqual(set(rss.parameters.keys()), set(rss_new.parameters.keys()))
        self.assertEqual(rss.parameter_distributions, rss_new.parameter_distributions)
        self.assertEqual(rss._environmental_variables, rss_new._environmental_variables)
        self.assertEqual(rss_new.parameters["c"].parameter_type, ParameterType.INT)

    def test_transform_experiment_data(self) -> None:
        parameterizations = [
            {"x": 2.2, "a": 2, "b": 10.0, "c": 10.0, "d": "r", "e": "q"},
            {"x": 1.0, "a": 1, "b": 1.0, "c": 100.0, "d": "q", "e": "z"},
            {"x": 1.2, "a": 2, "b": 100.0, "c": 1000.0, "d": "z", "e": "r"},
        ]
        experiment = get_experiment_with_observations(
            observations=[[1.0], [2.0], [3.0]],
            search_space=self.search_space,
            parameterizations=parameterizations,
        )
        experiment_data = extract_experiment_data(
            experiment=experiment, data_loader_config=DataLoaderConfig()
        )
        transformed_data = self.t.transform_experiment_data(
            experiment_data=deepcopy(experiment_data)
        )

        # Check that values in arm_data are transformed as expected.
        if self.t_class is ChoiceToNumericChoice:
            expected_values = zip(
                [2.2, 1.0, 1.2],
                [2, 1, 2],
                [10.0, 1.0, 100.0],
                [10.0, 100.0, 1000.0],
                [1, 0, 2],
                [1, 2, 0],
            )
        elif self.t_class is OrderedChoiceToIntegerRange:
            expected_values = zip(
                [2.2, 1.0, 1.2],
                [2, 1, 2],
                [1.0, 0.0, 2.0],
                [0.0, 1.0, 2.0],
                ["r", "q", "z"],
                ["q", "z", "r"],
            )
        else:
            raise NotImplementedError
        expected_arm_data = DataFrame(
            [
                {"x": x, "a": a, "b": b, "c": c, "d": d, "e": e}
                for x, a, b, c, d, e in expected_values
            ],
            index=experiment_data.arm_data.index,
        )
        assert_frame_equal(
            transformed_data.arm_data.drop(columns="metadata"), expected_arm_data
        )

        # Check that observation data is unchanged.
        assert_frame_equal(
            transformed_data.observation_data, experiment_data.observation_data
        )

        # Test with no parameters transformed.
        # Setting `encoded_parameters` directly to simplify testing.
        self.t.encoded_parameters = {}
        copy_experiment_data = deepcopy(experiment_data)
        transformed_data = self.t.transform_experiment_data(
            experiment_data=copy_experiment_data
        )
        # Arm data is same as before but it is not the same object.
        assert_frame_equal(transformed_data.arm_data, experiment_data.arm_data)
        self.assertIsNot(transformed_data.arm_data, copy_experiment_data.arm_data)
        # Observation data is the same object.
        assert_frame_equal(
            transformed_data.observation_data, experiment_data.observation_data
        )
        self.assertIs(
            transformed_data.observation_data, copy_experiment_data.observation_data
        )


class OrderedChoiceToIntegerRangeTransformTest(ChoiceEncodeTransformTest):
    t_class = OrderedChoiceToIntegerRange

    def setUp(self) -> None:
        super().setUp()
        # expected parameters after transform
        self.expected_transformed_params = {
            "x": 2.2,
            "a": 2,
            # float choice originally; transformed to int index.
            "b": 1,
            # float choice originally; transformed to int index.
            "c": 0,
            "d": "r",
        }

    def test_init(self) -> None:
        self.assertEqual(list(self.t.encoded_parameters.keys()), ["b", "c"])

    def test_transform_search_space(self) -> None:
        ss2 = deepcopy(self.search_space)
        ss2 = self.t.transform_search_space(ss2)

        for p in ("x", "a", "b", "c"):
            self.assertIsInstance(ss2.parameters[p], RangeParameter)
        self.assertIsInstance(ss2.parameters["d"], ChoiceParameter)
        self.assertEqual(ss2.parameters["x"].parameter_type, ParameterType.FLOAT)
        for p in ("a", "b", "c"):
            self.assertEqual(ss2.parameters[p].parameter_type, ParameterType.INT)
        self.assertEqual(ss2.parameters["d"].parameter_type, ParameterType.STRING)

        self.assertEqual(ss2.parameters["b"].lower, 0)
        self.assertEqual(ss2.parameters["b"].upper, 2)
        self.assertEqual(ss2.parameters["c"].lower, 0)
        self.assertEqual(ss2.parameters["c"].upper, 2)
        self.assertEqual(ss2.parameters["d"].values, ["q", "r", "z"])

    def test_transform_search_space_fidelity(self) -> None:
        # Ensure we error if we try to transform a fidelity parameter
        ss3 = SearchSpace(
            parameters=[
                ChoiceParameter(
                    "b",
                    parameter_type=ParameterType.FLOAT,
                    values=[1.0, 10.0, 100.0],
                    is_ordered=True,
                    is_fidelity=True,
                    target_value=100.0,
                )
            ]
        )
        t = OrderedChoiceToIntegerRange(search_space=ss3, observations=[])
        with self.assertRaises(ValueError):
            t.transform_search_space(ss3)

    def test_transform_search_space_with_different_values(self) -> None:
        # Parameter with unseen values.
        ss = SearchSpace(
            parameters=[
                ChoiceParameter(
                    name="b", parameter_type=ParameterType.FLOAT, values=[5.0, 10.0]
                )
            ]
        )
        with self.assertRaisesRegex(ValueError, "contains values that are not present"):
            self.t.transform_search_space(ss)
        # Parameter that maps to a non-contiguous range.
        ss = SearchSpace(
            parameters=[
                ChoiceParameter(
                    name="b", parameter_type=ParameterType.FLOAT, values=[1.0, 100.0]
                )
            ]
        )
        with self.assertRaisesRegex(ValueError, "does not span a contiguous range"):
            self.t.transform_search_space(ss)
        # Parameter that maps to a contiguous range not starting at 0.
        ss = SearchSpace(
            parameters=[
                ChoiceParameter(
                    name="b", parameter_type=ParameterType.FLOAT, values=[10.0, 100.0]
                )
            ]
        )
        t_ss = self.t.transform_search_space(ss)
        self.assertEqual(t_ss.parameters["b"].lower, 1)
        self.assertEqual(t_ss.parameters["b"].upper, 2)

    def test_deprecated_OrderedChoiceEncode(self) -> None:
        # Ensure we error if we try to transform a fidelity parameter
        ss3 = SearchSpace(
            parameters=[
                ChoiceParameter(
                    "b",
                    parameter_type=ParameterType.FLOAT,
                    values=[1.0, 10.0, 100.0],
                    is_ordered=True,
                    is_fidelity=True,
                    target_value=100.0,
                )
            ]
        )
        t = OrderedChoiceToIntegerRange(search_space=ss3, observations=[])
        t_deprecated = OrderedChoiceEncode(search_space=ss3, observations=[])
        self.assertEqual(t.__dict__, t_deprecated.__dict__)

    def test_hss_dependents_are_preserved(self) -> None:
        """
        Skip the HSS test. `OrderedChoiceToIntegerRange` cannot support hierarchical
        search spaces, because range parameters cannot have dependents.
        """


def normalize_values(values: Sized) -> list[float]:
    values = np.array(values, dtype=float)
    vmin, vmax = values.min(), values.max()
    if len(values) > 1:
        values = (values - vmin) / (vmax - vmin)
    return values.tolist()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.adapter.torch import TorchAdapter


# PairwiseAdapter was deprecated in Ax 1.1.0, so it should be reaped in Ax
# 1.2.0+
class PairwiseAdapter(TorchAdapter):
    # pyre-ignore[2]: Missing parameter annotations.
    def __init__(self, **kwargs) -> None:
        raise DeprecationWarning(
            "PairwiseAdapter is deprecated. Use TorchAdapter instead."
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from copy import deepcopy
from unittest import mock

import numpy as np
import numpy.typing as npt
from ax.adapter import Adapter
from ax.adapter.base import DataLoaderConfig
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.tests.test_relativize_transform import RelativizeDataTest
from ax.adapter.transforms.transform_to_new_sq import TransformToNewSQ
from ax.core.batch_trial import BatchTrial
from ax.core.observation import observations_from_data
from ax.exceptions.core import DataRequiredError
from ax.generators.base import Generator
from ax.utils.common.testutils import TestCase
from ax.utils.stats.statstools import relativize
from ax.utils.testing.core_stubs import (
    get_branin_data_batch,
    get_branin_experiment,
    get_branin_optimization_config,
    get_sobol,
)
from pandas.testing import assert_frame_equal
from pyre_extensions import assert_is_instance


class TransformToNewSQTest(RelativizeDataTest):
    # pyre-ignore [15]: `relativize_classes` overrides attribute
    # defined in `RelativizeDataTest` inconsistently. Type `List
    # [Type[TransformToNewSQ]]` is not a subtype of the
    # overridden attribute `List[Type[Transform]]`
    relativize_classes = [TransformToNewSQ]
    cases: list[tuple[type[Transform], list[tuple[npt.NDArray, npt.NDArray]]]] = [
        (
            TransformToNewSQ,
            [
                (
                    np.array([1.6, 10.0]),
                    np.array([[0.16, 0.0], [0.0, 0.2892562]]),
                ),
                (np.array([2.0, 5.0]), np.array([[0.1, 0.0], [0.0, 0.2]])),
                (np.array([1.0, 10.0]), np.array([[0.3, 0.0], [0.0, 0.4]])),
            ],
        )
    ]

    # these tests are defined by RelativizeDataTest, but it is irrelevant
    # for TransformToNewSQ, so we don't need to run it here.
    def test_bad_relativize(self) -> None:
        pass

    def test_transform_status_quos_always_zero(self) -> None:
        pass


class TransformToNewSQSpecificTest(TestCase):
    def setUp(self) -> None:
        self.exp = get_branin_experiment(
            with_batch=True,
            with_status_quo=True,
        )
        t = self.exp.trials[0]
        t.mark_running(no_runner_required=True)
        self.exp.attach_data(
            get_branin_data_batch(batch=assert_is_instance(t, BatchTrial))
        )
        t.mark_completed()
        self.data = self.exp.fetch_data()

        self._refresh_adapter()

    def _refresh_adapter(self) -> None:
        self.adapter = Adapter(
            search_space=self.exp.search_space,
            generator=Generator(),
            experiment=self.exp,
            data=self.exp.lookup_data(),
        )

    def test_adapter_without_status_quo_name(self) -> None:
        self.adapter._status_quo = None
        self.adapter._status_quo_name = None

        with self.assertRaisesRegex(
            DataRequiredError, "TransformToNewSQ requires status quo data."
        ):
            TransformToNewSQ(
                search_space=None,
                observations=[],
                adapter=self.adapter,
            )

    def test_transform_optimization_config(self) -> None:
        tf = TransformToNewSQ(
            search_space=None,
            observations=[],
            adapter=self.adapter,
        )
        oc = get_branin_optimization_config()
        new_oc = tf.transform_optimization_config(optimization_config=oc)
        self.assertIs(new_oc, oc)

    def test_untransform_outcome_constraints(self) -> None:
        tf = TransformToNewSQ(
            search_space=None,
            observations=[],
            adapter=self.adapter,
        )
        oc = get_branin_optimization_config()
        new_outcome_constraints = tf.untransform_outcome_constraints(
            outcome_constraints=oc.outcome_constraints
        )
        self.assertIs(new_outcome_constraints, oc.outcome_constraints)

    def test_custom_target_trial(self) -> None:
        tf = TransformToNewSQ(
            search_space=None,
            observations=[],
            adapter=self.adapter,
        )
        self.assertEqual(tf.default_trial_idx, 0)

        tf = TransformToNewSQ(
            search_space=None,
            observations=[],
            adapter=self.adapter,
            config={"target_trial_index": 1},
        )
        self.assertEqual(tf.default_trial_idx, 1)

    def test_single_trial_is_not_transformed(self) -> None:
        tf = TransformToNewSQ(
            search_space=None,
            observations=[],
            adapter=self.adapter,
        )
        obs = observations_from_data(
            experiment=self.exp,
            data=self.exp.lookup_data(),
        )[:1]
        obs2 = tf.transform_observations(observations=deepcopy(obs))
        self.assertEqual(obs, obs2)

    def test_target_trial_index(self) -> None:
        sobol = get_sobol(search_space=self.exp.search_space)
        self.exp.new_batch_trial(
            generator_run=sobol.gen(2), should_add_status_quo_arm=True
        )
        t = self.exp.trials[1]
        t = assert_is_instance(t, BatchTrial)
        t.mark_running(no_runner_required=True)
        self.exp.attach_data(
            get_branin_data_batch(batch=assert_is_instance(t, BatchTrial))
        )

        self._refresh_adapter()

        observations = observations_from_data(
            experiment=self.exp,
            data=self.exp.lookup_data(),
        )

        t = TransformToNewSQ(
            search_space=self.exp.search_space,
            observations=observations,
            adapter=self.adapter,
        )

        self.assertEqual(t.default_trial_idx, 1)

        with mock.patch(
            "ax.adapter.transforms.transform_to_new_sq.get_target_trial_index",
            return_value=0,
        ):
            t = TransformToNewSQ(
                search_space=self.exp.search_space,
                observations=observations,
                adapter=self.adapter,
            )

        self.assertEqual(t.default_trial_idx, 0)
        # test falling back to latest trial with SQ data
        with mock.patch(
            "ax.adapter.transforms.transform_to_new_sq.get_target_trial_index",
            return_value=10,
        ):
            t = TransformToNewSQ(
                search_space=self.exp.search_space,
                observations=observations,
                adapter=self.adapter,
            )

        self.assertEqual(t.default_trial_idx, 1)

    def test_transform_experiment_data(self) -> None:
        # Create two more trials with different SQ observations.
        sobol = get_sobol(search_space=self.exp.search_space)
        for sq_val in (2.0, 3.0):
            t = self.exp.new_batch_trial(
                generator_run=sobol.gen(2), should_add_status_quo_arm=True
            ).mark_completed(unsafe=True)
            data = get_branin_data_batch(batch=t)
            data.df.loc[(data.df["arm_name"] == "status_quo"), "mean"] = sq_val
            self.exp.attach_data(data=data)
        self._refresh_adapter()

        experiment_data = extract_experiment_data(
            experiment=self.exp, data_loader_config=DataLoaderConfig()
        )

        # Create the transform with trial 2 as the target.
        tf = TransformToNewSQ(
            search_space=None,
            adapter=self.adapter,
            config={"target_trial_index": 2},
        )
        transformed_data = tf.transform_experiment_data(
            experiment_data=deepcopy(experiment_data)
        )

        # Verify that status quo observations are dropped except for the target trial.
        tf_obs_data = transformed_data.observation_data
        self.assertEqual(
            tf_obs_data[
                tf_obs_data.index.get_level_values("arm_name")
                == self.adapter.status_quo_name
            ]
            .index.get_level_values("trial_index")
            .item(),
            2,  # target trial index from the config.
        )
        # Verify that data from the target trial is not transformed.
        target_trial_data = experiment_data.observation_data.loc[2]
        transformed_target_trial_data = transformed_data.observation_data.loc[2]
        assert_frame_equal(target_trial_data, transformed_target_trial_data)

        # Check that the data for trials 0 and 1 are transformed correctly.
        sq_data_target = self.adapter.status_quo_data_by_trial[2]
        for t_idx in (0, 1):
            sq_data = self.adapter.status_quo_data_by_trial[t_idx]
            # Get the data for the non-sq arms.
            trial_data = experiment_data.observation_data.loc[t_idx]
            trial_data = trial_data[
                trial_data.index.get_level_values("arm_name")
                != self.adapter.status_quo_name
            ]
            # Relativize the data with respect to the SQ for the trial.
            means_rel, sems_rel = relativize(
                means_t=trial_data["mean", "branin"],
                sems_t=trial_data["sem", "branin"],
                mean_c=sq_data.means[0],
                sem_c=sq_data.covariance[0, 0] ** 0.5,
                as_percent=False,
                control_as_constant=tf.control_as_constant,
            )
            # Derelativize using target SQ.
            target_mean = sq_data_target.means[0]
            abs_target_mean = np.abs(target_mean)
            means = means_rel * abs_target_mean + target_mean
            sems = sems_rel * abs_target_mean
            self.assertTrue(
                np.allclose(
                    means,
                    transformed_data.observation_data.loc[t_idx]["mean", "branin"],
                )
            )
            self.assertTrue(
                np.allclose(
                    sems, transformed_data.observation_data.loc[t_idx]["sem", "branin"]
                )
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import logging
import tempfile
from datetime import datetime
from itertools import product
from logging import WARNING
from math import pi
from time import monotonic
from typing import Literal
from unittest.mock import patch

import numpy as np
import torch
from ax.adapter.factory import get_sobol
from ax.adapter.registry import Generators
from ax.benchmark.benchmark import (
    _get_oracle_value_of_params,
    benchmark_multiple_problems_methods,
    benchmark_one_method_problem,
    benchmark_replication,
    compute_baseline_value_from_sobol,
    compute_score_trace,
    get_benchmark_orchestrator_options,
    get_benchmark_result_from_experiment_and_gs,
    get_benchmark_result_with_cumulative_steps,
    get_best_parameters,
    get_opt_trace_by_steps,
    get_oracle_experiment_from_params,
    run_optimization_with_orchestrator,
)
from ax.benchmark.benchmark_method import BenchmarkMethod
from ax.benchmark.benchmark_problem import (
    BenchmarkProblem,
    get_continuous_search_space,
    get_moo_opt_config,
    get_soo_opt_config,
)
from ax.benchmark.benchmark_result import BenchmarkResult
from ax.benchmark.benchmark_runner import BenchmarkRunner
from ax.benchmark.benchmark_test_functions.synthetic import IdentityTestFunction
from ax.benchmark.methods.modular_botorch import (
    get_sobol_botorch_modular_acquisition,
    get_sobol_mbm_generation_strategy,
)
from ax.benchmark.methods.sobol import (
    get_sobol_benchmark_method,
    get_sobol_generation_strategy,
)
from ax.benchmark.problems.registry import get_benchmark_problem
from ax.benchmark.problems.synthetic.from_botorch import (
    create_problem_from_botorch,
    get_augmented_branin_problem,
)
from ax.benchmark.testing.benchmark_stubs import (
    get_async_benchmark_method,
    get_async_benchmark_problem,
    get_discrete_search_space,
    get_moo_surrogate,
    get_multi_objective_benchmark_problem,
    get_single_objective_benchmark_problem,
    get_soo_surrogate,
)

from ax.core.experiment import Experiment
from ax.core.map_data import MapData
from ax.early_stopping.strategies.threshold import ThresholdEarlyStoppingStrategy
from ax.generation_strategy.external_generation_node import ExternalGenerationNode
from ax.generation_strategy.generation_strategy import (
    GenerationNode,
    GenerationStrategy,
)
from ax.generation_strategy.generator_spec import GeneratorSpec
from ax.service.utils.orchestrator_options import TrialType
from ax.storage.json_store.load import load_experiment
from ax.storage.json_store.save import save_experiment
from ax.utils.common.logger import get_logger
from ax.utils.common.mock import mock_patch_method_original
from ax.utils.common.testutils import TestCase

from ax.utils.testing.core_stubs import get_experiment_with_observations
from ax.utils.testing.mock import mock_botorch_optimize
from botorch.acquisition.knowledge_gradient import qKnowledgeGradient
from botorch.acquisition.logei import qLogNoisyExpectedImprovement
from botorch.acquisition.multi_objective.logei import (
    qLogNoisyExpectedHypervolumeImprovement,
)
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.gp_regression import SingleTaskGP
from botorch.optim.optimize import optimize_acqf

from botorch.test_functions.synthetic import Branin
from pyre_extensions import assert_is_instance, none_throws


class TestBenchmark(TestCase):
    def benchmark_replication(
        self,
        problem: BenchmarkProblem,
        method: BenchmarkMethod,
        seed: int,
        strip_runner_before_saving: bool = True,
    ) -> BenchmarkResult:
        """
        Run benchmark_replication with logs set to WARNING.

        Suppresses voluminous INFO logs from the orchestrator.
        """
        return benchmark_replication(
            problem=problem,
            method=method,
            seed=seed,
            strip_runner_before_saving=strip_runner_before_saving,
            orchestrator_logging_level=WARNING,
        )

    def run_optimization_with_orchestrator(
        self,
        problem: BenchmarkProblem,
        method: BenchmarkMethod,
        seed: int,
    ) -> Experiment:
        """
        Run run_optimization_with_orchestrator with logs set to WARNING.

        Suppresses voluminous INFO logs from the orchestrator.
        """
        return run_optimization_with_orchestrator(
            problem=problem,
            method=method,
            seed=seed,
            orchestrator_logging_level=WARNING,
        )

    @mock_botorch_optimize
    def test_batch(self) -> None:
        batch_size = 5

        problem = get_benchmark_problem("ackley4", num_trials=2)
        for sequential in [False, True]:
            with self.subTest(sequential=sequential):
                batch_method_joint = get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    batch_size=batch_size,
                    distribute_replications=False,
                    model_gen_kwargs={
                        "model_gen_options": {
                            "optimizer_kwargs": {"sequential": sequential}
                        }
                    },
                    num_sobol_trials=1,
                )
                with patch(
                    "ax.generators.torch.botorch_modular.acquisition.optimize_acqf",
                    wraps=optimize_acqf,
                ) as mock_optimize_acqf:
                    self.run_optimization_with_orchestrator(
                        problem=problem, method=batch_method_joint, seed=0
                    )
                mock_optimize_acqf.assert_called_once()
                self.assertEqual(
                    mock_optimize_acqf.call_args.kwargs["sequential"], sequential
                )
                self.assertEqual(mock_optimize_acqf.call_args.kwargs["q"], batch_size)

    def _test_storage(self, map_data: bool) -> None:
        problem = get_async_benchmark_problem(map_data=map_data)
        method = get_async_benchmark_method()
        res = self.benchmark_replication(problem=problem, method=method, seed=0)
        # Experiment is not in storage yet
        self.assertTrue(res.experiment is not None)
        self.assertEqual(res.experiment_storage_id, None)
        experiment = res.experiment

        # test saving to temporary file
        with tempfile.NamedTemporaryFile(mode="w", delete=True, suffix=".json") as f:
            save_experiment(none_throws(res.experiment), f.name)
            res.experiment_storage_id = f.name
            res.experiment = None
            self.assertIsNone(res.experiment)
            self.assertEqual(res.experiment_storage_id, f.name)

            # load it back
            experiment = load_experiment(f.name)
            self.assertEqual(experiment, experiment)

    def test_storage(self) -> None:
        self._test_storage(map_data=False)
        self._test_storage(map_data=True)

    def test_replication_sobol_synthetic(self) -> None:
        method = get_sobol_benchmark_method(distribute_replications=False)
        problems = [
            get_single_objective_benchmark_problem(),
            get_benchmark_problem("jenatton", num_trials=6),
        ]
        for problem in problems:
            res = self.benchmark_replication(problem=problem, method=method, seed=0)

            self.assertEqual(
                problem.num_trials, len(none_throws(res.experiment).trials)
            )
            self.assertTrue(np.isfinite(res.score_trace).all())
            self.assertTrue(np.all(res.score_trace <= 100))
            experiment = none_throws(res.experiment)
            self.assertIn(f"{problem.name}|Sobol", experiment.name)
            self.assertEqual(experiment.search_space, problem.search_space)
            self.assertEqual(
                experiment.optimization_config, problem.optimization_config
            )

    def test_compute_score_trace(self) -> None:
        opt_trace = np.array([1, 0, -1, 2, float("nan"), 4])

        with self.subTest("Higher is better"):
            optimal_value = 5
            baseline_value = 1
            expected_trace = np.array([0, -25, -50, 25, float("nan"), 75.0])
            trace = compute_score_trace(
                optimization_trace=opt_trace,
                baseline_value=baseline_value,
                optimal_value=optimal_value,
            )
            self.assertTrue(np.array_equal(trace, expected_trace, equal_nan=True))

        with self.subTest("Lower is better"):
            optimal_value = -1
            baseline_value = 0
            expected_trace = np.array([-100, 0, 100, -200, float("nan"), -400])
            trace = compute_score_trace(
                optimization_trace=opt_trace,
                baseline_value=baseline_value,
                optimal_value=optimal_value,
            )
            self.assertTrue(np.array_equal(trace, expected_trace, equal_nan=True))

    def test_replication_sobol_surrogate(self) -> None:
        method = get_sobol_benchmark_method(distribute_replications=False)

        # This is kind of a weird setup - these are "surrogates" that use a Branin
        # synthetic function. The idea here is to test the machinery around the
        # surrogate benchmarks without having to actually load a surrogate model
        # of potentially non-neglible size.
        for name, problem in [
            ("soo", get_soo_surrogate()),
            ("moo", get_moo_surrogate()),
        ]:
            with self.subTest(name, problem=problem):
                res = self.benchmark_replication(problem=problem, method=method, seed=0)

                self.assertEqual(
                    problem.num_trials,
                    len(none_throws(res.experiment).trials),
                )

                self.assertTrue(np.isfinite(res.score_trace).all())
                self.assertTrue(np.all(res.score_trace <= 100))

    def _test_replication_async(self, map_data: bool) -> None:
        """
        The test function is the identity function, higher is better, observed
        to be noiseless, and the same at every point on the trajectory. And the
        generation strategy deterministically produces
        candidates with values 0, 1, 2, .... So if the trials complete in order,
        the optimization trace should be 0, 1, 2, .... If the trials complete
        out of order, the traces should track the argmax of the completion
        order.

        Args:
            map_data: If True, the test function produces time-series data with
                just one step, so behavior is the same as when map_data=False.
        """
        method = get_async_benchmark_method()

        complete_out_of_order_runtimes = {
            0: 2,
            1: 1,
            2: 3,
            3: 1,
        }
        step_runtime_fns = {
            "All complete at different times": lambda params: params["x0"] * 3,
            "Trials complete immediately": lambda params: 0,
            "Trials complete at same time": lambda params: 1,
            "Complete out of order": lambda params: complete_out_of_order_runtimes[
                params["x0"]
            ],
        }

        # First case:
        # Time   | trial 0 | trial 1 | trial 2 | trial 3
        #  t=0   |   .     |   .     |         |
        #  t=1-2 |         |   .     |   .     |
        #  t=3-6 |         |         |   .     |    .
        #  t=7-12|         |         |         |    .

        # Second case:
        # Time   | trial 0 | trial 1 | trial 2 | trial 3
        #  t=0   |   .     |   .     |         |
        #  t=1   |   .     |         |   .     |
        #  t=2   |         |         |   .     |    .
        #  t=3   |         |         |   .     |
        expected_start_times = {
            "All complete at different times": [0, 0, 1, 3],
            "Trials complete immediately": [0, 0, 1, 1],
            # Without MapData, completing after 0 seconds (second case) has the
            # same effect as completing after 1 second (third case), because a
            # new trial can't start until the next time increment.
            # With MapData, trials complete at the same times as without
            # MapData, but an extra step accrues in the third case.
            "Trials complete at same time": [0, 0, 1, 1],
            "Complete out of order": [0, 0, 1, 2],
        }
        expected_pending_in_each_gen = {
            "All complete at different times": [[None], [0], [1], [2]],
            "Trials complete immediately": [[None], [0], [None], [2]],
            "Trials complete at same time": [[None], [0], [None], [2]],
            "Complete out of order": [[None], [0], [0], [2]],
        }
        # When two trials complete at the same time, the inference trace uses
        # data from both to get the best point, and repeats it.
        expected_traces = {
            "All complete at different times": [0, 1, 2, 3],
            # 0 and 1 complete at the same time, as do 2 and 3
            "Trials complete immediately": [1, 3],
            "Trials complete at same time": [1, 3],
            "Complete out of order": [1, 1, 3, 3],
        }
        expected_costs = {
            "All complete at different times": [0, 3, 7, 12],
            "Trials complete immediately": [0, 1],
            "Trials complete at same time": [1, 2],
            "Complete out of order": [1, 2, 3, 4],
        }
        expected_backend_simulator_time = {
            "All complete at different times": 12,
            "Trials complete immediately": 2,
            "Trials complete at same time": 2,
            "Complete out of order": 4,
        }

        for case_name, step_runtime_fn in step_runtime_fns.items():
            with self.subTest(case_name, step_runtime_fn=step_runtime_fn):
                problem = get_async_benchmark_problem(
                    map_data=map_data,
                    step_runtime_fn=step_runtime_fn,
                )

                with mock_patch_method_original(
                    mock_path=(
                        "ax.benchmark.testing.benchmark_stubs."
                        "ExternalGenerationNode._gen"
                    ),
                    original_method=ExternalGenerationNode._gen,
                ) as mock_gen:
                    result = self.benchmark_replication(
                        problem=problem,
                        method=method,
                        seed=0,
                        strip_runner_before_saving=False,
                    )
                pending_in_each_gen = [
                    [
                        elt[0].trial_index
                        for elt in call_kwargs.get("pending_observations").values()
                    ]
                    for _, call_kwargs in mock_gen.call_args_list
                ]
                self.assertEqual(
                    pending_in_each_gen,
                    expected_pending_in_each_gen[case_name],
                    case_name,
                )

                experiment = none_throws(result.experiment)
                runner = assert_is_instance(experiment.runner, BenchmarkRunner)
                backend_simulator = none_throws(
                    runner.simulated_backend_runner
                ).simulator
                self.assertEqual(
                    backend_simulator.time,
                    expected_backend_simulator_time[case_name],
                    msg=case_name,
                )
                completed_trials = backend_simulator.state().completed
                self.assertEqual(len(completed_trials), 4)

                params_dict = {
                    idx: trial.arms[0].parameters
                    for idx, trial in experiment.trials.items()
                }
                expected_runtimes = {
                    idx: step_runtime_fn(params=params)
                    for idx, params in params_dict.items()
                }
                for trial_index, expected_start_time in enumerate(
                    expected_start_times[case_name]
                ):
                    self.assertEqual(trial_index, params_dict[trial_index]["x0"])
                    expected_runtime = expected_runtimes[trial_index]
                    self.assertEqual(
                        backend_simulator.get_sim_trial_by_index(
                            trial_index=trial_index
                        ).__dict__,
                        {
                            "trial_index": trial_index,
                            "sim_runtime": expected_runtime,
                            "sim_start_time": expected_start_time,
                            "sim_queued_time": expected_start_time,
                            "sim_completed_time": expected_start_time
                            + expected_runtime,
                        },
                        f"Failure for trial {trial_index} with {case_name}",
                    )
                self.assertFalse(np.isnan(result.inference_trace).any())
                self.assertEqual(
                    result.inference_trace.tolist(),
                    expected_traces[case_name],
                    msg=case_name,
                )
                self.assertEqual(
                    result.oracle_trace.tolist(),
                    expected_traces[case_name],
                    msg=case_name,
                )
                self.assertEqual(
                    result.cost_trace.tolist(),
                    expected_costs[case_name],
                    msg=case_name,
                )
                if map_data:
                    data = assert_is_instance(experiment.lookup_data(), MapData)
                    self.assertEqual(len(data.df), 4, msg=case_name)
                    self.assertEqual(len(data.map_df), 4, msg=case_name)

                # Check trial start and end times
                start_of_time = datetime.fromtimestamp(0)
                created_times = [
                    (trial.time_created - start_of_time).total_seconds()
                    for i, trial in experiment.trials.items()
                ]
                self.assertEqual(created_times, expected_start_times[case_name])
                queued_times = [
                    (
                        none_throws(experiment.trials[i].time_run_started)
                        - start_of_time
                    ).total_seconds()
                    for i in range(4)
                ]
                self.assertEqual(queued_times, expected_start_times[case_name])
                completed_times = [
                    (
                        none_throws(experiment.trials[i].time_completed) - start_of_time
                    ).total_seconds()
                    for i in range(4)
                ]
                expected_completed_times = [
                    expected_start_times[case_name][i] + expected_runtimes[i]
                    for i in range(4)
                ]
                self.assertEqual(completed_times, expected_completed_times)

    def test_replication_async(self) -> None:
        self._test_replication_async(map_data=False)
        self._test_replication_async(map_data=True)

    def test_run_optimization_with_orchestrator(self) -> None:
        method = get_async_benchmark_method()
        problem = get_async_benchmark_problem(
            map_data=True,
        )

        # Test logging
        logger = get_logger("utils.testing.backend_simulator")

        with self.subTest("Logs produced if level is DEBUG"):
            with self.assertLogs(level=logging.DEBUG, logger=logger):
                experiment = run_optimization_with_orchestrator(
                    problem=problem,
                    method=method,
                    seed=0,
                    orchestrator_logging_level=logging.DEBUG,
                )
            runner = assert_is_instance(experiment.runner, BenchmarkRunner)
            self.assertFalse(
                none_throws(runner.simulated_backend_runner).simulator._verbose_logging
            )

        with self.subTest("Logs not produced by default"), self.assertNoLogs(
            level=logging.INFO, logger=logger
        ), self.assertNoLogs(logger=logger):
            run_optimization_with_orchestrator(
                problem=problem,
                method=method,
                seed=0,
            )

    def test_early_stopping(self) -> None:
        """
        Test early stopping with a deterministic generation strategy and ESS
        that stops if the objective exceeds 0.5 when their progression ("t") hits 2,
        which happens when 3 steps have passed (t=[0, 1, 2]).

        Each arm produces values equaling the trial index everywhere on the
        progression, so Trials 1, 2, and 3 will stop early, and trial 0 will not.

        t=0-2: Trials 0 and 1 run.
        t=3: Trial 1 stops early. Trial 2 gets added to "_queued", and then to
            "_running", with a queued time of 3 and a sim_start_time of 4.
        t=4: Trials 0 and 2 run.
        t=5: Trial 0 completes. Trial 2 runs.
        t=6: Trials 2 runs. Trial 3 starts with a sim_queued_time of 5 and a
            sim_start_time of 5.
        t=7: Trial 2 stops early. Trial 3 runs.
        t=8-9: Trial 3 runs by itself then gets stopped early.
        """
        min_progression = 2
        progression_length_if_not_stopped = 5
        early_stopping_strategy = ThresholdEarlyStoppingStrategy(
            metric_threshold=0.5,
            min_progression=min_progression,
            min_curves=0,
        )

        method = get_async_benchmark_method(
            early_stopping_strategy=early_stopping_strategy
        )

        problem = get_async_benchmark_problem(
            map_data=True,
            n_steps=progression_length_if_not_stopped,
            lower_is_better=True,
        )
        experiment = self.run_optimization_with_orchestrator(
            problem=problem, method=method, seed=0
        )
        data = assert_is_instance(experiment.lookup_data(), MapData)
        expected_n_steps = {
            0: progression_length_if_not_stopped,
            # stopping after step=2, so 3 steps (0, 1, 2) have passed
            **{i: min_progression + 1 for i in range(1, 4)},
        }

        grouped = data.map_df.groupby("trial_index")
        self.assertEqual(
            dict(grouped["step"].count()),
            expected_n_steps,
        )
        for trial_index, sub_df in grouped:
            self.assertEqual(
                sub_df["step"].tolist(),
                list(range(expected_n_steps[trial_index])),
                msg=f"Trial {trial_index}",
            )
        self.assertEqual(
            dict(grouped["step"].max()),
            {
                0: progression_length_if_not_stopped - 1,
                **{i: min_progression for i in range(1, 4)},
            },
        )
        simulator = none_throws(
            assert_is_instance(
                experiment.runner, BenchmarkRunner
            ).simulated_backend_runner
        ).simulator
        trials = {
            trial_index: none_throws(simulator.get_sim_trial_by_index(trial_index))
            for trial_index in range(4)
        }
        start_times = {
            trial_index: sim_trial.sim_start_time
            for trial_index, sim_trial in trials.items()
        }
        expected_start_times = {
            0: 0,
            1: 0,
            2: 4,
            3: 5,
        }
        self.assertEqual(start_times, expected_start_times)

        with self.subTest("max_pending_trials = 1"):
            method = get_async_benchmark_method(
                early_stopping_strategy=early_stopping_strategy,
                max_pending_trials=1,
            )
            experiment = self.run_optimization_with_orchestrator(
                problem=problem, method=method, seed=0
            )
            simulated_backend_runner = assert_is_instance(
                experiment.runner, BenchmarkRunner
            ).simulated_backend_runner
            self.assertIsNotNone(simulated_backend_runner)
            expected_start_times = {
                0: 0,
                1: 5,  # Early-stopped at t=8
                2: 9,  # Early-stopped at t=11
                3: 13,  # Early-stopped at t=16
            }
            simulator = none_throws(simulated_backend_runner).simulator
            trials = {
                trial_index: none_throws(simulator.get_sim_trial_by_index(trial_index))
                for trial_index in range(4)
            }
            start_times = {
                trial_index: sim_trial.sim_start_time
                for trial_index, sim_trial in trials.items()
            }
            self.assertEqual(start_times, expected_start_times)
            map_df = assert_is_instance(experiment.lookup_data(), MapData).map_df
            max_run = map_df.groupby("trial_index")["step"].max().to_dict()
            self.assertEqual(max_run, {0: 4, 1: 2, 2: 2, 3: 2})

    def test_replication_variable_runtime(self) -> None:
        method = get_async_benchmark_method(max_pending_trials=1)
        for map_data in [False, True]:
            with self.subTest(map_data=map_data):
                problem = get_async_benchmark_problem(
                    map_data=map_data,
                    step_runtime_fn=lambda params: params["x0"] + 1,
                )
                experiment = self.run_optimization_with_orchestrator(
                    problem=problem, method=method, seed=0
                )
                simulated_backend_runner = assert_is_instance(
                    experiment.runner, BenchmarkRunner
                ).simulated_backend_runner
                self.assertIsNotNone(simulated_backend_runner)
                expected_start_times = {
                    0: 0,
                    1: 1,
                    2: 3,
                    3: 6,
                }
                simulator = none_throws(simulated_backend_runner).simulator
                trials = {
                    trial_index: none_throws(
                        simulator.get_sim_trial_by_index(trial_index)
                    )
                    for trial_index in range(4)
                }
                start_times = {
                    trial_index: sim_trial.sim_start_time
                    for trial_index, sim_trial in trials.items()
                }
                self.assertEqual(start_times, expected_start_times)

    @mock_botorch_optimize
    def _test_replication_with_inference_value(
        self, batch_size: int, report_inference_value_as_trace: bool
    ) -> None:
        seed = 1
        method = get_sobol_botorch_modular_acquisition(
            model_cls=SingleTaskGP,
            acquisition_cls=qLogNoisyExpectedImprovement,
            distribute_replications=False,
            num_sobol_trials=3,
            batch_size=batch_size,
        )

        num_trials = 4
        problem = get_single_objective_benchmark_problem(
            num_trials=num_trials,
            report_inference_value_as_trace=report_inference_value_as_trace,
            noise_std=100.0,
        )
        res = self.benchmark_replication(problem=problem, method=method, seed=seed)
        # The inference trace could coincide with the oracle trace, but it won't
        # happen in this example with high noise and a seed
        self.assertEqual(
            np.equal(res.inference_trace, res.optimization_trace).all(),
            report_inference_value_as_trace,
        )
        self.assertEqual(
            np.equal(res.oracle_trace, res.optimization_trace).all(),
            not report_inference_value_as_trace,
        )

        self.assertEqual(res.optimization_trace.shape, (problem.num_trials,))
        self.assertTrue((res.inference_trace >= res.oracle_trace).all())

    def test_replication_with_inference_value(self) -> None:
        for batch_size, report_inference_value_as_trace in product(
            [1, 2], [False, True]
        ):
            with self.subTest(
                batch_size=batch_size,
                report_inference_value_as_trace=report_inference_value_as_trace,
            ):
                self._test_replication_with_inference_value(
                    batch_size=batch_size,
                    report_inference_value_as_trace=report_inference_value_as_trace,
                )

        with self.assertRaisesRegex(
            NotImplementedError,
            "Inference trace is not supported for MOO",
        ):
            get_multi_objective_benchmark_problem(report_inference_value_as_trace=True)

    @mock_botorch_optimize
    def test_replication_mbm(self) -> None:
        for method, problem, expected_name in [
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    distribute_replications=True,
                ),
                get_benchmark_problem(
                    "constrained_gramacy_observed_noise", num_trials=6
                ),
                "MBM::SingleTaskGP_qLogNEI",
            ),
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    distribute_replications=False,
                ),
                get_single_objective_benchmark_problem(
                    observe_noise_sd=True, num_trials=6
                ),
                "MBM::SingleTaskGP_qLogNEI",
            ),
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    distribute_replications=False,
                ),
                get_single_objective_benchmark_problem(
                    observe_noise_sd=True, num_trials=6
                ),
                "MBM::SingleTaskGP_qLogNEI",
            ),
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedHypervolumeImprovement,
                    distribute_replications=False,
                ),
                get_multi_objective_benchmark_problem(
                    observe_noise_sd=True, num_trials=6
                ),
                "MBM::SingleTaskGP_qLogNEHVI",
            ),
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SaasFullyBayesianSingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    distribute_replications=False,
                ),
                get_multi_objective_benchmark_problem(num_trials=6),
                "MBM::SAAS_qLogNEI",
            ),
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qKnowledgeGradient,
                    distribute_replications=False,
                ),
                get_single_objective_benchmark_problem(
                    observe_noise_sd=False, num_trials=6
                ),
                "MBM::SingleTaskGP_qKnowledgeGradient",
            ),
            (
                get_sobol_botorch_modular_acquisition(
                    model_cls=SingleTaskGP,
                    acquisition_cls=qLogNoisyExpectedImprovement,
                    distribute_replications=False,
                ),
                get_augmented_branin_problem(fidelity_or_task="fidelity"),
                "MBM::SingleTaskGP_qLogNEI",
            ),
        ]:
            with self.subTest(method=method, problem=problem):
                res = self.benchmark_replication(problem=problem, method=method, seed=0)
                self.assertEqual(
                    problem.num_trials,
                    len(none_throws(res.experiment).trials),
                )
                self.assertTrue(np.all(res.score_trace <= 100))
                self.assertEqual(method.name, method.generation_strategy.name)
                self.assertEqual(method.name, expected_name)

    def test_replication_moo_sobol(self) -> None:
        problem = get_multi_objective_benchmark_problem()

        res = self.benchmark_replication(
            problem=problem,
            method=get_sobol_benchmark_method(distribute_replications=False),
            seed=0,
        )

        self.assertEqual(
            problem.num_trials,
            len(none_throws(res.experiment).trials),
        )
        self.assertEqual(
            problem.num_trials * 2,
            len(none_throws(res.experiment).fetch_data().df),
        )

        self.assertTrue(np.all(res.score_trace <= 100))
        self.assertEqual(len(res.cost_trace), problem.num_trials)
        self.assertEqual(len(res.inference_trace), problem.num_trials)
        # since inference trace is not supported for MOO, it should be all NaN
        self.assertTrue(np.isnan(res.inference_trace).all())

    def test_benchmark_one_method_problem(self) -> None:
        problem = get_single_objective_benchmark_problem()
        method = get_sobol_benchmark_method(distribute_replications=False)
        with self.assertNoLogs(level="INFO"):
            agg = benchmark_one_method_problem(
                problem=problem,
                method=method,
                seeds=(0, 1),
                orchestrator_logging_level=WARNING,
            )

        self.assertEqual(len(agg.results), 2)
        self.assertTrue(
            all(
                len(none_throws(result.experiment).trials) == problem.num_trials
                for result in agg.results
            ),
            "All experiments must have 4 trials",
        )

        for col in ["mean", "P25", "P50", "P75"]:
            self.assertTrue((agg.score_trace[col] <= 100).all())

    @mock_botorch_optimize
    def test_benchmark_multiple_problems_methods(self) -> None:
        problems = [get_single_objective_benchmark_problem(num_trials=6)]
        methods = [
            get_sobol_benchmark_method(distribute_replications=False),
            get_sobol_botorch_modular_acquisition(
                model_cls=SingleTaskGP,
                acquisition_cls=qLogNoisyExpectedImprovement,
                distribute_replications=False,
            ),
        ]
        with self.assertNoLogs(level="INFO"):
            aggs = benchmark_multiple_problems_methods(
                problems=problems,
                methods=methods,
                seeds=(0, 1),
                orchestrator_logging_level=WARNING,
            )

        self.assertEqual(len(aggs), 2)
        for agg in aggs:
            for col in ["mean", "P25", "P50", "P75"]:
                self.assertTrue((agg.score_trace[col] <= 100).all())

    def test_timeout(self) -> None:
        problem = create_problem_from_botorch(
            test_problem_class=Branin,
            test_problem_kwargs={},
            num_trials=1000,  # Unachievable num_trials
            baseline_value=100,
        )

        generation_strategy = get_sobol_mbm_generation_strategy(
            model_cls=SingleTaskGP,
            acquisition_cls=qLogNoisyExpectedImprovement,
            num_sobol_trials=1000,  # Ensures we don't use BO
        )
        timeout_seconds = 2.0
        method = BenchmarkMethod(
            generation_strategy=generation_strategy,
            timeout_hours=timeout_seconds / 3600,
        )

        # Each replication will have a different number of trials

        start = monotonic()
        with self.assertLogs("ax.service.orchestrator", level="ERROR") as cm:
            result = benchmark_one_method_problem(
                problem=problem,
                method=method,
                seeds=(0, 1),
                orchestrator_logging_level=WARNING,
            )
        elapsed = monotonic() - start
        self.assertGreater(elapsed, timeout_seconds)
        self.assertTrue(any("Optimization timed out" in output for output in cm.output))

        # Test the traces get composited correctly. The AggregatedResult's traces
        # should be the length of the shortest trace in the BenchmarkResults
        min_num_trials = min(len(res.optimization_trace) for res in result.results)
        self.assertEqual(len(result.optimization_trace), min_num_trials)
        self.assertEqual(len(result.score_trace), min_num_trials)

    def test_replication_with_generation_node(self) -> None:
        method = BenchmarkMethod(
            name="Sobol Generation Node",
            generation_strategy=GenerationStrategy(
                nodes=[
                    GenerationNode(
                        node_name="Sobol",
                        generator_specs=[
                            GeneratorSpec(
                                Generators.SOBOL, model_kwargs={"deduplicate": True}
                            )
                        ],
                    )
                ]
            ),
        )
        problem = get_single_objective_benchmark_problem()
        with self.assertNoLogs(logger=get_logger("ax.core.experiment"), level="INFO"):
            res = self.benchmark_replication(problem=problem, method=method, seed=0)

        # Check that logger level has been reset
        self.assertEqual(get_logger("ax.core.experiment").level, logging.INFO)
        self.assertEqual(problem.num_trials, len(none_throws(res.experiment).trials))
        self.assertFalse(np.isnan(res.score_trace).any())

    def test_get_oracle_experiment_from_params(self) -> None:
        problem = create_problem_from_botorch(
            test_problem_class=Branin,
            test_problem_kwargs={},
            num_trials=5,
        )
        # first is near optimum
        near_opt_params = {"x0": -pi, "x1": 12.275}
        other_params = {"x0": 0.5, "x1": 0.5}
        unbatched_experiment = get_oracle_experiment_from_params(
            problem=problem,
            dict_of_dict_of_params={0: {"0": near_opt_params}, 1: {"1": other_params}},
        )
        self.assertEqual(len(unbatched_experiment.trials), 2)
        self.assertTrue(
            all(t.status.is_completed for t in unbatched_experiment.trials.values())
        )
        self.assertTrue(
            all(len(t.arms) == 1 for t in unbatched_experiment.trials.values())
        )
        df = unbatched_experiment.fetch_data().df
        self.assertAlmostEqual(df["mean"].iloc[0], Branin._optimal_value, places=5)

        batched_experiment = get_oracle_experiment_from_params(
            problem=problem,
            dict_of_dict_of_params={0: {"0_0": near_opt_params, "0_1": other_params}},
        )
        self.assertEqual(len(batched_experiment.trials), 1)
        self.assertEqual(len(batched_experiment.trials[0].arms), 2)
        df = batched_experiment.fetch_data().df
        self.assertAlmostEqual(df["mean"].iloc[0], Branin._optimal_value, places=5)

        # Test empty inputs
        experiment = get_oracle_experiment_from_params(
            problem=problem, dict_of_dict_of_params={}
        )
        self.assertEqual(len(experiment.trials), 0)

        with self.assertRaisesRegex(ValueError, "trial with no arms"):
            get_oracle_experiment_from_params(
                problem=problem, dict_of_dict_of_params={0: {}}
            )

    def _test_multi_fidelity_or_multi_task(
        self, fidelity_or_task: Literal["fidelity", "task"]
    ) -> None:
        """
        Args:
            fidelity_or_task: "fidelity" or "task"
        """
        problem = get_augmented_branin_problem(fidelity_or_task=fidelity_or_task)
        params = {"x0": 1.0, "x1": 0.0, "x2": 0.0}
        at_target = assert_is_instance(
            Branin()
            .evaluate_true(torch.tensor([1.0, 0.0], dtype=torch.double).unsqueeze(0))
            .item(),
            float,
        )
        oracle_experiment = get_oracle_experiment_from_params(
            problem=problem, dict_of_dict_of_params={0: {"0": params}}
        )
        self.assertAlmostEqual(
            oracle_experiment.fetch_data().df["mean"].iloc[0],
            at_target,
        )
        # first term: (-(b - 0.1) * (1 - x3)  + c - r)^2
        # low-fidelity: (-b - 0.1 + c - r)^2
        # high-fidelity: (-b + c - r)^2
        t = -5.1 / (4 * pi**2) + 5 / pi - 6
        expected_change = (t + 0.1) ** 2 - t**2
        self.assertAlmostEqual(
            problem.test_function.evaluate_true(params=params).item(),
            at_target + expected_change,
        )

    def test_multi_fidelity_or_multi_task(self) -> None:
        self._test_multi_fidelity_or_multi_task(fidelity_or_task="fidelity")
        self._test_multi_fidelity_or_multi_task(fidelity_or_task="task")

    def test_get_benchmark_orchestrator_options(self) -> None:
        for include_sq, batch_size in product((False, True), (1, 2)):
            method = BenchmarkMethod(
                generation_strategy=get_sobol_mbm_generation_strategy(
                    model_cls=SingleTaskGP, acquisition_cls=qLogNoisyExpectedImprovement
                ),
                distribute_replications=False,
                max_pending_trials=2,
                batch_size=batch_size,
            )
            orchestrator_options = get_benchmark_orchestrator_options(
                method=method, include_sq=include_sq
            )
            self.assertEqual(orchestrator_options.max_pending_trials, 2)
            self.assertEqual(orchestrator_options.init_seconds_between_polls, 0)
            self.assertEqual(orchestrator_options.min_seconds_before_poll, 0)
            self.assertEqual(orchestrator_options.batch_size, batch_size)
            self.assertEqual(
                orchestrator_options.run_trials_in_batches, method.run_trials_in_batches
            )
            self.assertEqual(
                orchestrator_options.early_stopping_strategy,
                method.early_stopping_strategy,
            )
            self.assertEqual(
                orchestrator_options.trial_type,
                TrialType.BATCH_TRIAL
                if include_sq or batch_size > 1
                else TrialType.TRIAL,
            )
            self.assertEqual(
                orchestrator_options.status_quo_weight, 1.0 if include_sq else 0.0
            )

    def test_replication_with_status_quo(self) -> None:
        method = BenchmarkMethod(
            name="Sobol", generation_strategy=get_sobol_generation_strategy()
        )
        problem = get_single_objective_benchmark_problem(
            status_quo_params={"x0": 0.0, "x1": 0.0}
        )
        experiment = self.run_optimization_with_orchestrator(
            problem=problem, method=method, seed=0
        )

        self.assertEqual(problem.num_trials, len(experiment.trials))
        for t in experiment.trials.values():
            self.assertEqual(len(t.arms), 2, msg=f"Trial index: {t.index}")
            self.assertEqual(
                sum(a.name == "status_quo" for a in t.arms),
                1,
                msg=f"Trial index: {t.index}",
            )

    def test_compute_baseline_value_from_sobol(self) -> None:
        """
        In this setting, every point from 0-4 will be evaluated,
        and it will produce outcomes 0-4.
        """
        search_space = get_discrete_search_space(n_values=5)
        test_function = IdentityTestFunction()

        with self.subTest("SOO, lower is better"):
            opt_config = get_soo_opt_config(outcome_names=test_function.outcome_names)
            result = compute_baseline_value_from_sobol(
                optimization_config=opt_config,
                search_space=search_space,
                test_function=test_function,
                n_repeats=1,
            )
            self.assertEqual(result, 0)

        with self.subTest("SOO, MapData"):
            map_test_function = IdentityTestFunction(n_steps=2)
            map_opt_config = get_soo_opt_config(
                outcome_names=test_function.outcome_names, use_map_metric=True
            )
            result = compute_baseline_value_from_sobol(
                optimization_config=map_opt_config,
                search_space=search_space,
                test_function=map_test_function,
                n_repeats=1,
            )
            self.assertEqual(result, 0)

        with self.subTest("SOO, higher is better"):
            opt_config = get_soo_opt_config(
                outcome_names=test_function.outcome_names, lower_is_better=False
            )
            result = compute_baseline_value_from_sobol(
                optimization_config=opt_config,
                search_space=search_space,
                test_function=test_function,
                n_repeats=1,
            )
            self.assertEqual(result, 4)

        moo_test_function = IdentityTestFunction(outcome_names=["foo", "bar"])
        with self.subTest("MOO"):
            moo_opt_config = get_moo_opt_config(
                outcome_names=moo_test_function.outcome_names, ref_point=[5, 5]
            )
            result = compute_baseline_value_from_sobol(
                optimization_config=moo_opt_config,
                search_space=search_space,
                test_function=moo_test_function,
                n_repeats=1,
            )
            # (5-0) * (5-0)
            self.assertEqual(result, 25)

    def test_get_oracle_value_of_params(self) -> None:
        problem = get_augmented_branin_problem(fidelity_or_task="fidelity")
        # params are not at target value
        params = {"x0": 1.0, "x1": 0.0, "x2": 0.0}
        inference_value = _get_oracle_value_of_params(params=params, problem=problem)
        oracle_params = {"x0": 1.0, "x1": 0.0, "x2": 1.0}
        self.assertEqual(
            inference_value,
            problem.test_function.evaluate_true(params=oracle_params).item(),
        )

    def test_get_opt_trace_by_cumulative_epochs(self) -> None:
        # Time  | trial 0 | trial 1 | trial 2 | trial 3 | new steps
        #  t=0  |   ..    |  start  |         |         |   0, 0
        #  t=1  |         |    .    |  start  |         |   1
        #  t=2  |         |    ..   |         |  start  |   1
        #  t=3  |         |         |    .    |         |   2
        #  t=4  |         |         |         |         |
        #  t=5  |         |         |    ..   |    .    |   3, 2

        # ...
        problem = get_async_benchmark_problem(
            map_data=True,
            n_steps=2,
            # Ensure we don't have two finishing at the same time, for
            # determinism
            step_runtime_fn=lambda params: params["x0"] * (1 - 0.01 * params["x0"]),
        )
        method = get_async_benchmark_method()

        with self.subTest("Without early stopping"):
            experiment = self.run_optimization_with_orchestrator(
                problem=problem, method=method, seed=0
            )
            new_opt_trace = get_opt_trace_by_steps(experiment=experiment)

            self.assertEqual(
                list(new_opt_trace), [0.0, 0.0, 1.0, 1.0, 2.0, 3.0, 3.0, 3.0]
            )

        with self.subTest("With early stopping"):
            es_method = get_async_benchmark_method(
                early_stopping_strategy=ThresholdEarlyStoppingStrategy(
                    metric_threshold=10.0, min_progression=0, min_curves=2
                )
            )
            experiment = self.run_optimization_with_orchestrator(
                problem=problem, method=es_method, seed=0
            )
            new_opt_trace = get_opt_trace_by_steps(experiment=experiment)
            self.assertEqual(list(new_opt_trace), [0.0, 0.0, 1.0, 1.0, 2.0, 3.0])

        method = get_sobol_benchmark_method(distribute_replications=False)
        with self.subTest("MOO"):
            problem = get_multi_objective_benchmark_problem()

            experiment = self.run_optimization_with_orchestrator(
                problem=problem, method=method, seed=0
            )
            with self.assertRaisesRegex(
                NotImplementedError, "only supported for single objective"
            ):
                get_opt_trace_by_steps(experiment=experiment)

        with self.subTest("Constrained"):
            problem = get_benchmark_problem("constrained_gramacy_observed_noise")
            experiment = self.run_optimization_with_orchestrator(
                problem=problem, method=method, seed=0
            )
            with self.assertRaisesRegex(
                NotImplementedError,
                "not supported for problems with outcome constraints",
            ):
                get_opt_trace_by_steps(experiment=experiment)

    def test_get_benchmark_result_with_cumulative_steps(self) -> None:
        """See test_get_opt_trace_by_cumulative_epochs for more info."""
        problem = get_async_benchmark_problem(
            map_data=True,
            n_steps=2,
            # Ensure we don't have two finishing at the same time, for
            # determinism
            step_runtime_fn=lambda params: params["x0"] * (1 - 0.01 * params["x0"]),
        )
        method = get_async_benchmark_method()
        result = self.benchmark_replication(problem=problem, method=method, seed=0)
        transformed = get_benchmark_result_with_cumulative_steps(
            result=result,
            optimal_value=problem.optimal_value,
            baseline_value=problem.baseline_value,
        )
        map_df = assert_is_instance(
            none_throws(result.experiment).lookup_data(), MapData
        ).map_df
        self.assertEqual(len(map_df), len(transformed.optimization_trace))
        self.assertEqual(
            list(transformed.optimization_trace),
            [0.0, 0.0, 1.0, 1.0, 2.0, 3.0, 3.0, 3.0],
        )
        self.assertTrue(np.isnan(transformed.oracle_trace).all())
        self.assertTrue(np.isnan(transformed.inference_trace).all())
        self.assertEqual(transformed.score_trace.max(), result.score_trace.max())
        self.assertLessEqual(transformed.score_trace.min(), result.score_trace.min())

    def test_get_best_parameters(self) -> None:
        """
        Whether this produces the correct values is tested more thoroughly in
        other tests such as `test_replication_with_inference_value` and
        `test_get_inference_trace_from_params`.  Setting up an experiment with
        data and trials without just running a benchmark is a pain, so in those
        tests, we just run a benchmark.
        """
        gs = get_sobol_generation_strategy()

        search_space = get_continuous_search_space(bounds=[(0, 1)])
        moo_config = get_moo_opt_config(outcome_names=["a", "b"], ref_point=[0, 0])
        experiment = Experiment(
            name="test",
            is_test=True,
            search_space=search_space,
            optimization_config=moo_config,
        )

        with self.subTest("MOO not supported"), self.assertRaisesRegex(
            NotImplementedError, "Please use `get_pareto_optimal_parameters`"
        ):
            get_best_parameters(experiment=experiment, generation_strategy=gs)

        soo_config = get_soo_opt_config(outcome_names=["a"])
        with self.subTest("Empty experiment"):
            result = get_best_parameters(
                experiment=experiment.clone_with(optimization_config=soo_config),
                generation_strategy=gs,
            )
            self.assertIsNone(result)

        with self.subTest("All constraints violated"):
            experiment = get_experiment_with_observations(
                observations=[[1, -1], [2, -1]],
                constrained=True,
            )
            best_point = get_best_parameters(
                experiment=experiment, generation_strategy=gs
            )
            self.assertIsNone(best_point)

        with self.subTest("No completed trials"):
            experiment = get_experiment_with_observations(observations=[])
            sobol_generator = get_sobol(search_space=experiment.search_space)
            for _ in range(3):
                trial = experiment.new_trial(generator_run=sobol_generator.gen(n=1))
                trial.run()
            best_point = get_best_parameters(
                experiment=experiment, generation_strategy=gs
            )
            self.assertIsNone(best_point)

        experiment = get_experiment_with_observations(
            observations=[[1], [2]], constrained=False
        )
        with self.subTest("Working case"):
            best_point = get_best_parameters(
                experiment=experiment, generation_strategy=gs
            )
            self.assertEqual(best_point, experiment.trials[1].arms[0].parameters)

        with self.subTest("Trial indices"):
            best_point = get_best_parameters(
                experiment=experiment, generation_strategy=gs, trial_indices=[0]
            )
            self.assertEqual(best_point, experiment.trials[0].arms[0].parameters)

    def test_get_benchmark_result_from_experiment_and_gs(self) -> None:
        problem = get_single_objective_benchmark_problem()
        method = BenchmarkMethod(
            name="Sobol", generation_strategy=get_sobol_generation_strategy()
        )
        seed = 0
        result = self.benchmark_replication(
            problem=problem, method=method, seed=seed, strip_runner_before_saving=False
        )

        result2 = get_benchmark_result_from_experiment_and_gs(
            experiment=none_throws(result.experiment),
            generation_strategy=method.generation_strategy,
            problem=problem,
            seed=seed,
            strip_runner_before_saving=False,
        )
        # Idempotency
        self.assertEqual(result, result2)
        # Runner not stripped
        self.assertIsNotNone(none_throws(result2.experiment).runner)
        self.assertEqual(result2.seed, seed)

        with self.subTest("runner stripped"):
            result = get_benchmark_result_from_experiment_and_gs(
                experiment=none_throws(result.experiment),
                generation_strategy=method.generation_strategy,
                problem=problem,
                seed=3,
            )
            self.assertIsNone(none_throws(result.experiment).runner)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Sequence

from ax.core.map_metric import MapMetric

from ax.core.objective import MultiObjective, Objective, ScalarizedObjective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.outcome_constraint import (
    ComparisonOp,
    ObjectiveThreshold,
    OutcomeConstraint,
    ScalarizedOutcomeConstraint,
)
from ax.core.parameter_constraint import ParameterConstraint
from ax.exceptions.core import UserInputError
from ax.utils.common.string_utils import sanitize_name, unsanitize_name
from pyre_extensions import assert_is_instance, none_throws
from sympy.core.add import Add
from sympy.core.expr import Expr
from sympy.core.mul import Mul
from sympy.core.relational import GreaterThan, LessThan
from sympy.core.symbol import Symbol
from sympy.core.sympify import sympify


def optimization_config_from_string(
    objective_str: str, outcome_constraint_strs: Sequence[str] | None = None
) -> OptimizationConfig:
    """
    Create an OptimizationConfig from objective and outcome constraint strings.

    Note that outcome constraints may not be placed on the objective metric except in
    the multi-objective case where they will be converted to objective thresholds.
    """

    objective = parse_objective(objective_str=objective_str)

    if outcome_constraint_strs is not None:
        outcome_constraints = [
            parse_outcome_constraint(constraint_str=constraint_str)
            for constraint_str in outcome_constraint_strs
        ]
    else:
        outcome_constraints = None

    if isinstance(objective, MultiObjective):
        # Convert OutcomeConstraints to ObjectiveThresholds if relevant
        objective_metric_names = {metric.name for metric in objective.metrics}
        true_outcome_constraints = []
        objective_thresholds: list[ObjectiveThreshold] = []
        for outcome_constraint in outcome_constraints or []:
            if (
                not isinstance(outcome_constraint, ScalarizedOutcomeConstraint)
                and outcome_constraint.metric.name in objective_metric_names
            ):
                objective_thresholds.append(
                    ObjectiveThreshold(
                        metric=outcome_constraint.metric,
                        bound=outcome_constraint.bound,
                        relative=outcome_constraint.relative,
                        op=outcome_constraint.op,
                    )
                )
            else:
                true_outcome_constraints.append(outcome_constraint)

        return MultiObjectiveOptimizationConfig(
            objective=objective,
            outcome_constraints=true_outcome_constraints,
            objective_thresholds=objective_thresholds,
        )

    # Ensure that outcome constraints are not placed on the objective metric
    objective_metric_names = {metric.name for metric in objective.metrics}
    for outcome_constraint in outcome_constraints or []:
        if outcome_constraint.metric.name in objective_metric_names:
            raise UserInputError(
                "Outcome constraints may not be placed on the objective metric "
                f"except in the multi-objective case, found {objective_str} and "
                f"{outcome_constraint_strs}"
            )

    return OptimizationConfig(
        objective=objective,
        outcome_constraints=outcome_constraints,
    )


def parse_parameter_constraint(constraint_str: str) -> ParameterConstraint:
    """
    Parse a parameter constraint string into a ParameterConstraint object using SymPy.
    Currently only supports linear constraints of the form "a * x + b * y >= k" or
    "a * x + b * y <= k".
    """
    coefficient_dict = _extract_coefficient_dict_from_inequality(
        inequality_str=constraint_str
    )

    # Iterate through the coefficients to extract the parameter names and weights and
    # the bound
    constraint_dict = {}
    bound = 0
    for term, coefficient in coefficient_dict.items():
        if term.is_symbol:
            constraint_dict[unsanitize_name(term.name)] = coefficient
        elif term.is_number:
            # Invert because we are "moving" the bound to the right hand side
            bound = -1 * coefficient
        else:
            raise UserInputError(
                "Only linear inequality parameter constraints are supported, found "
                f"{constraint_str}"
            )

    return ParameterConstraint(constraint_dict=constraint_dict, bound=bound)


def parse_objective(objective_str: str) -> Objective:
    """
    Parse an objective string into an Objective object using SymPy.

    Currently only supports linear objectives of the form "a * x + b * y" and tuples of
    linear objectives.
    """
    # Parse the objective string into a SymPy expression
    expression = sympify(sanitize_name(objective_str))

    if isinstance(expression, tuple):  # Multi-objective
        return MultiObjective(
            objectives=[
                _create_single_objective(expression=term) for term in expression
            ]
        )

    return _create_single_objective(expression=assert_is_instance(expression, Expr))


def parse_outcome_constraint(constraint_str: str) -> OutcomeConstraint:
    """
    Parse an outcome constraint string into an OutcomeConstraint object using SymPy.
    Currently only supports linear constraints of the form "a * x + b * y >= k" or
    "a * x + b * y <= k".

    To indicate a relative constraint (i.e. performance relative to some baseline)
    multiply your bound by "baseline". For example "qps >= 0.95 * baseline" will
    constrain such that the QPS is at least 95% of the baseline arm's QPS.
    """
    coefficient_dict = _extract_coefficient_dict_from_inequality(
        inequality_str=constraint_str
    )

    # Iterate through the coefficients to extract the parameter names and weights and
    # the bound
    constraint_dict: dict[str, float] = {}
    bound = 0
    is_relative = False
    for term, coefficient in coefficient_dict.items():
        if term.is_symbol:
            if term.name == "baseline":
                # Invert because we are "moving" the bound to the right hand side
                bound = -1 * coefficient
                is_relative = True
            else:
                constraint_dict[term.name] = coefficient
        elif term.is_number:
            # Invert because we are "moving" the bound to the right hand side
            bound = -1 * coefficient
        else:
            raise UserInputError(
                "Only linear outcome constraints are supported, found "
                f"{constraint_str}"
            )

    if len(constraint_dict) == 1:
        term, coefficient = next(iter(constraint_dict.items()))

        return OutcomeConstraint(
            metric=MapMetric(name=unsanitize_name(term)),
            op=ComparisonOp.LEQ if coefficient > 0 else ComparisonOp.GEQ,
            bound=bound / coefficient,
            relative=is_relative,
        )

    names, coefficients = zip(*constraint_dict.items())
    return ScalarizedOutcomeConstraint(
        metrics=[MapMetric(name=unsanitize_name(name)) for name in names],
        op=ComparisonOp.LEQ,
        weights=[*coefficients],
        bound=bound,
        relative=is_relative,
    )


def _create_single_objective(expression: Expr) -> Objective:
    """
    Create an Objective or ScalarizedObjective from a linear SymPy expression.

    All expressions are assumed to represent maximization objectives.
    """

    # If the expression is a just a Symbol it represents a single metric objective
    if isinstance(expression, Symbol):
        return Objective(
            metric=MapMetric(
                name=unsanitize_name(str(expression.name)), lower_is_better=False
            ),
            minimize=False,
        )

    # If the expression is a Mul it likely represents a single metric objective but
    # some additional validation is required
    if isinstance(expression, Mul):
        symbol, *other_symbols = expression.free_symbols
        if len(other_symbols) > 0:
            raise UserInputError(
                f"Only linear objectives are supported, found {expression}."
            )

        # Since the objectives 1 * loss and 2 * loss are equivalent, we can just use
        # the sign from the coefficient rather than its value
        minimize = bool(
            none_throws(expression.as_coefficient(assert_is_instance(symbol, Expr))) < 0
        )

        return Objective(
            metric=MapMetric(
                name=unsanitize_name(str(symbol)), lower_is_better=minimize
            ),
            minimize=minimize,
        )

    # If the expression is an Add it represents a scalarized objective
    elif isinstance(expression, Add):
        names, coefficients = zip(*expression.as_coefficients_dict().items())
        return ScalarizedObjective(
            metrics=[MapMetric(name=unsanitize_name(str(name))) for name in names],
            weights=[float(coefficient) for coefficient in coefficients],
            minimize=False,
        )

    raise UserInputError(f"Only linear objectives are supported, found {expression}.")


def _extract_coefficient_dict_from_inequality(
    inequality_str: str,
) -> dict[Symbol, float]:
    """
    Use SymPy to parse a string into an inequality, invert if necessary to enforce a
    less-than relationship, move all terms to the left side, and return the
    coefficients as a dictionary. This is useful for parsing parameter and outcome
    constraints.
    """
    # Parse the constraint string into a SymPy inequality
    inequality = sympify(sanitize_name(inequality_str))

    # Check the SymPy object is a valid inequality
    if not isinstance(inequality, GreaterThan | LessThan):
        raise UserInputError(f"Expected an inequality, found {inequality_str}")

    # Move all terms to the left side of the inequality and invert if necessary to
    # enforce a less-than relationship
    if isinstance(inequality, LessThan):
        expression = inequality.lhs - inequality.rhs
    else:
        expression = inequality.rhs - inequality.lhs

    return {
        key: float(value) for key, value in expression.as_coefficients_dict().items()
    }



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.analysis.plotly.arm_effects import ArmEffectsPlot
from ax.analysis.plotly.bandit_rollout import BanditRollout
from ax.analysis.plotly.cross_validation import CrossValidationPlot
from ax.analysis.plotly.marginal_effects import MarginalEffectsPlot
from ax.analysis.plotly.parallel_coordinates import ParallelCoordinatesPlot
from ax.analysis.plotly.plotly_analysis import PlotlyAnalysisCard
from ax.analysis.plotly.progression import ProgressionPlot
from ax.analysis.plotly.scatter import ScatterPlot
from ax.analysis.plotly.sensitivity import SensitivityAnalysisPlot
from ax.analysis.plotly.surface.contour import ContourPlot
from ax.analysis.plotly.surface.slice import SlicePlot
from ax.analysis.plotly.top_surfaces import TopSurfacesAnalysis

__all__ = [
    "ArmEffectsPlot",
    "BanditRollout",
    "ContourPlot",
    "CrossValidationPlot",
    "MarginalEffectsPlot",
    "ParallelCoordinatesPlot",
    "PlotlyAnalysisCard",
    "ProgressionPlot",
    "ScatterPlot",
    "SensitivityAnalysisPlot",
    "SlicePlot",
    "TopSurfacesAnalysis",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.adapter.registry import Generators
from ax.benchmark.benchmark_method import BenchmarkMethod
from ax.generation_strategy.generation_strategy import (
    GenerationStep,
    GenerationStrategy,
)


def get_sobol_generation_strategy() -> GenerationStrategy:
    return GenerationStrategy(
        name="Sobol",
        steps=[
            GenerationStep(generator=Generators.SOBOL, num_trials=-1),
        ],
    )


def get_sobol_benchmark_method(
    distribute_replications: bool,
    batch_size: int = 1,
) -> BenchmarkMethod:
    return BenchmarkMethod(
        generation_strategy=get_sobol_generation_strategy(),
        batch_size=batch_size,
        distribute_replications=distribute_replications,
    )
